/* --------------------------------------------------------------------------
 * DatUtil - Written by Logiqx (http://www.logiqx.com)
 *
 * A simple little utility for converting, cleansing and comparing ROM manager
 * dats (e.g. ClrMamePro, RomCenter). Also supports ROM lists generated by
 * emulators such as Retrocade, Callus and System 16 so that dats can be checked
 * for missing/spurious ROMs.
 *
 * Additional formats can easily be added if you have know a little C.
 * -------------------------------------------------------------------------- */

#define DATUTIL_VERSION "v1.11"
#define DATUTIL_DATE "28 June 2001"

/* --- The standard includes --- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <dirent.h>
#include <sys/stat.h>
#include <zlib.h>

/* --- Although DJGPP has a getopt function, MinGW, CygWin and MSVC don't --- */

#include "getopt.c"	// Same functions as in unistd library

/* --- Definitions for MAME's unzip routines --- */

#include "mame/unzip.h"

/* --- My type definitions and macros --- */

#include "datutil.h"


/* --- The Dat formats and ROM list formats supported --- */

LOAD_SAVE_FORMAT(romcenter_dat, "RomCenter Data", "ROMCENTER", ROMCENT_SUPPORT|AUTO_SORT)
LOAD_SAVE_FORMAT(mame_listinfo, "MAME Listinfo", "LISTINFO", FULL_SUPPORT)
LOAD_SAVE_FORMAT(romcenter2_dat, "RomCenter 2 Data", "ROMCENTER2", ROMCENT2_SUPPORT)
SAVE_ONLY_FORMAT(title_list, "Title List", "TITLELIST", AUTO_SORT)
LOAD_SAVE_FORMAT(mess_listinfo, "MESS Listinfo", "LISTINFO", FULL_SUPPORT)
LOAD_ONLY_FORMAT(raine_gameinfo, "RAINE Gameinfo", FULL_SUPPORT)
LOAD_ONLY_FORMAT(shark_listinfo, "Shark Listinfo", SIZE_SUPPORT|ALT_ROM_SUPPORT)
LOAD_ONLY_FORMAT(callus_gamelist, "Callus Gamelist", PARENT_SUPPORT)
LOAD_ONLY_FORMAT(retrocade_romlist, "Retrocade Romlist", 0)
LOAD_ONLY_FORMAT(system16_gcs, "System16 GCS", PARENT_SUPPORT|ONE_GAME_ONLY|SAMPLE_SUPPORT)
LOAD_ONLY_FORMAT(jff_txt, "Just For Fun Text", SIZE_SUPPORT)
LOAD_ONLY_FORMAT(sparcade_romlist, "Sparcade Romlist", ALT_ROM_SUPPORT)
LOAD_ONLY_FORMAT(sparcade_rat, "Sparcade RAT File", ALT_ROM_SUPPORT)
LOAD_ONLY_FORMAT(jas_romlist, "JAS Romlist", SIZE_SUPPORT)
LOAD_ONLY_FORMAT(mimic_gcs, "Mimic GCS", 0)
LOAD_SAVE_FORMAT(calice_drv, "Calice Driver", "CALICE", PARENT_SUPPORT|AUTO_SORT)
LOAD_ONLY_FORMAT(clrmame_log, "ClrMamePro Log", FULL_SUPPORT)

const struct DatFormat *formats[] =
{
	FORMAT(mame_listinfo)
	FORMAT(romcenter2_dat)
	FORMAT(romcenter_dat)
	FORMAT(mess_listinfo)
	FORMAT(raine_gameinfo)
	FORMAT(shark_listinfo)
	FORMAT(callus_gamelist)
	FORMAT(retrocade_romlist)
	FORMAT(system16_gcs)
	FORMAT(jff_txt)
	/* FORMAT(sparcade_romlist) - Not enabled. See reasons in source */
	FORMAT(sparcade_rat)
	FORMAT(jas_romlist)
	FORMAT(mimic_gcs)
	FORMAT(calice_drv)
	FORMAT(clrmame_log)
	FORMAT(title_list)
	0
};

/* --------------------------------------------------------------------------
 * The main() function just works out what the user wants to do then calls the
 * appropriate function (convert_dat or compare_dats).
 *
 * Uses the getopt() function from unistd to interpret command line options.
 * -------------------------------------------------------------------------- */

char refname[LONG_NAME_LENGTH];
char fullname[LONG_NAME_LENGTH];
char version[LONG_NAME_LENGTH];
char comment[LONG_NAME_LENGTH];
char author[LONG_NAME_LENGTH];
char merging[10];
char zipping[10];

int main(int argc, char **argv)
{
	/* --- For getopt function --- */

	extern char *optarg;
	extern int optind;

	/* --- Options chosen by user ---*/

	struct options *useropts;

	/* --- Misc variables ---*/

	int c, errflg = 0;

	/* --- Credits! --- */

	printf("===============================================================================\n");
#ifdef LONG_NAMES
	printf("DatUtil %s (%s) - long name version\n", DATUTIL_VERSION, DATUTIL_DATE);
#else
	printf("DatUtil %s (%s)\n", DATUTIL_VERSION, DATUTIL_DATE);
#endif
	printf("Written by Logiqx (http://www.logiqx.com)\n");
	printf("===============================================================================\n");

	/* --- Allocate memory for user options --- */

	if (!(useropts=calloc(1, sizeof(struct options))))
	{
		printf("Unable to allocate options memory\n");
		errflg++;
	}

	/* --- Get the options specified on the command line --- */

	while (!errflg && (c = getopt(argc, argv, "g:co:a:f:srmA:V:C:R:F:M:Z:")) != EOF)
	switch (c)
	{
		case 'g':
			useropts->game=optarg;
			break;
		case 'c':
			useropts->options|=GAME_AND_CLONES;
			break;
		case 'o':
			useropts->outfile=optarg;
			useropts->outmode="w";
			break;
		case 'a':
			useropts->outfile=optarg;
			useropts->outmode="a";
			break;
		case 'f':
			UPPER(optarg);
			useropts->format=optarg;
			break;
		case 's':
			useropts->options|=SORT_GAMES;
			break;
		case 'r':
			useropts->options|=REMOVE_CLONES;
			break;
		case 'm':
			useropts->options|=FIX_MERGING;
			break;
		case 'A':
			useropts->rc_author=optarg;
			break;
		case 'V':
			useropts->rc_version=optarg;
			break;
		case 'C':
			useropts->rc_comment=optarg;
			break;
		case 'R':
			useropts->rc_refname=optarg;
			break;
		case 'F':
			useropts->rc_fullname=optarg;
			break;
		case 'M':
			useropts->rc_merging=optarg;
			break;
		case 'Z':
			useropts->rc_zipping=optarg;
			break;
		case '?':
			errflg++;														/* User wants help! */
	}

	/* --- The user must specify either one or two dat files --- */

	if (argc-optind<1 || argc-optind>2)
		errflg++;

	/* --- Display the help page if required --- */

	if (errflg)
	{
		printf("Usage 1: datutil [-g <game> [-c]] [-o|a <outfile>] [-f <format>] [-s] [-r] [-m]\n");
		printf("                 [-A <author>] [-V <version>] [-C <comment>] [-R <refname>]\n");
		printf("                 [-F <fullname>] [-M <merging>] [-Z <zipping>] <infile>\n\n");
		printf("Converts datfile to a different format (the output format can be specified by\nusing -f). To specify the output filename use the -o option (or -a for append).\nIndividual games can be converted using the -g option (using -c will also\nconvert it's clones). If required, the -s option can be used to sort the games\nand the -r option can be used to remove clones. The -A, -V, -C, -R, -F, -M and\n-Z options are used to specify header contents when saving in ROMCenter format.\n\nValid save formats are listinfo, romcenter, romcenter2 and titlelist.\n\n");
		printf("Usage 2: datutil [-g <game> [-c]] [-o|a <logfile>] [-r] <datfile1> <datfile2>\n\n");
		printf("Compares two datfiles and reports any differences. To specify the name for the\nlogfile use the -o option (or -a for append). Individual games can be compared\nusing the -g option (using -c will also also compare it's clones). When\ncomparing dats, the -r option will remove clones from the comparison.\n");
	}
	else
	{
		if (!(useropts->outmode))
			useropts->outmode="w";

		if (optind==argc-1)
		{
			/* --- If the user didn't specify the outmode, use overwrite --- */

			if (!useropts->outfile)
				useropts->outfile="datutil.dat";

			useropts->datname1=argv[optind];

			errflg=convert_dat(useropts);
		}
		else
		{
			if (!useropts->outfile)
				useropts->outfile="datutil.log";

			useropts->datname1=argv[optind];
			useropts->datname2=argv[optind+1];

			errflg=compare_dats(useropts);
		}
	}

	if (useropts) free(useropts);

	/* --- All done --- */

	exit(errflg);
}

/* --------------------------------------------------------------------------
 * Dat conversion / extraction of individual game (and clones if required).
 * Basically just identifies the source dat, loads it into memory then
 * saves it back out again in the appropriate format.
 * -------------------------------------------------------------------------- */

int convert_dat(struct options *useropts)
{
	struct dat dat1;
	int errflg = 0;
	int i, save_format=-1;

	init_dat(&dat1, useropts->datname1);

	/* --- If dat was identified and contains sufficient details to save... --- */

	if (dat1.flags & SAVE_SUPPORT)
	{
		/* --- Load the dat into memory --- */

		LOAD_DAT(dat1)

		/* --- If successful --- */

		if (!errflg)
		{
			display_dat_info(&dat1);

			/* --- Search for first save format different to original --- */
			for (i=0;formats[i];i++)
				if (save_format==-1 && formats[i]->save!=dat1.save)
					save_format=i;

			/* --- Overide with original save format if saving individual game --- */
			/* --- ...but only if the original format can save! --- */
			if (useropts->game && dat1.save)
				for (i=0;formats[i];i++)
					if (formats[i]->save==dat1.save)
						save_format=i;

			/* --- Again, overide if format parameter was specified --- */
			if (useropts->format)
			{
				for (errflg=1, i=0;formats[i];i++)
				{
					if (!strcmp(useropts->format, formats[i]->option))
					{
						save_format=i;
						errflg=0;
					}
				}
				if (errflg)
					printf("Unknown save format '%s'\n", useropts->format);
			}

			/* --- Sort if user asked for it (or if autosorting is required) --- */
			if (((useropts->options) & SORT_GAMES) || (formats[save_format]->flags & AUTO_SORT))
				qsort(dat1.games, dat1.num_games, sizeof(struct game), parent_sort_function);

			/* --- Now we've decided the format, save the dat --- */
			(formats[save_format]->save)(&dat1, useropts);
		}

		release_dat(&dat1);

		return(0);
	}
	else if (dat1.flags)
	{
		printf("%s does not contain sufficient information for conversion.\n", dat1.descr);
		return(1);
	}

	return(0);
}

/* --------------------------------------------------------------------------
 * Dat comparison. There are two algorithms here... both sort the games and
 * compare the resultant games lists, the difference is how ROMs are compared:
 *
 * 1) If neither dat is a Shark listinfo (which allows several names for each
 *    ROM) it sorts the ROMs and compares them in the same way as with games.
 * 2) If either dat is Shark listinfo then for each ROM in a game, every
 *    possible name is compared against all names for the ROMs in the other dat!
 * -------------------------------------------------------------------------- */

/* --- The maximum number of ROMs for one game --- */

#define MAX_GAME_ROMS 256

int compare_dats(struct options *useropts)
{
	struct dat dat1, dat2;
	struct game *games1, *games2;

	FILE *out;

	struct rom *rom_ptr;
	struct alt_rom *alt_rom_ptr;
	struct rom roms1[MAX_GAME_ROMS], roms2[MAX_GAME_ROMS];
	struct alt_rom alt_roms1[MAX_GAME_ROMS], alt_roms2[MAX_GAME_ROMS];
	int found, sizediff;
	char size1[SIZE_LENGTH+1], size2[SIZE_LENGTH+1];
	int num_roms1, num_roms2;

	int errflg = 0;

	int i=0, j=0, gamediff;
	int k, l, m, n, romdiff;
	int total_diffs=0;
	int bad_size1=0, bad_size2=0;
	int bad_romof1=0, bad_romof2=0;
	int bad_merge1=0, bad_merge2=0;

	/* --- Identify the dat types... --- */

	init_dat(&dat1, useropts->datname1);
	init_dat(&dat2, useropts->datname2);

	/* --- Load them in, display their summary and sort the games --- */

	if (dat1.flags && dat2.flags)
		LOAD_DAT(dat1)
	else
		errflg++;

	if (!errflg)
	{
		display_dat_info(&dat1);
		qsort(dat1.games, dat1.num_games, sizeof(struct game), game_sort_function);
	}

	if (!errflg && dat2.flags)
		LOAD_DAT(dat2)
	else
		errflg++;

	if (!errflg)
	{
		display_dat_info(&dat2);
		qsort(dat2.games, dat2.num_games, sizeof(struct game), game_sort_function);
	}
	
	/* --- Open the log file --- */

	if (!errflg && !(out=fopen(useropts->outfile, useropts->outmode)))
	{
		printf("Error opening %s\n", useropts->outfile);
		return(1);
	}

	/* --- If neither dat uses alternative ROM names, use sort method 1 --- */

	if (!errflg && !((dat1.flags & ALT_ROM_SUPPORT) | (dat2.flags & ALT_ROM_SUPPORT)))
	{
		games1=dat1.games; games2=dat2.games;

		/* --- For all games --- */

		while (i<dat1.num_games || j<dat2.num_games)
		{
			/* --- If reached the end of either dat there is a game missing --- */

			if (i>=dat1.num_games) gamediff=1;
			if (j>=dat2.num_games) gamediff=-1;

			/* --- Else compare the game names --- */

			if (i<dat1.num_games && j<dat2.num_games)
				gamediff=strcmp(games1[i].name, games2[j].name);

			/* --- If the game names are the same then compare their details --- */

			if (gamediff==0)
			{
				/* --- Compare their parent details --- */

				if ((dat1.flags & PARENT_SUPPORT) && (dat2.flags & PARENT_SUPPORT) &&
					strcmp(games1[i].parent_name, games2[j].parent_name))
				{
					fprintf(out, "%s: **game** has different parents ('%s' and '%s')\n", games1[i].name, games1[i].parent_name, games2[j].parent_name);
					total_diffs++;
				}

				/* --- Compare their romof details --- */

				if ((dat1.flags & ROMOF_SUPPORT) && (dat2.flags & ROMOF_SUPPORT) &&
					strcmp(games1[i].romof, games2[j].romof))
				{
					fprintf(out, "%s: **game** has different romof values ('%s' and '%s')\n", games1[i].name, games1[i].romof, games2[j].romof);
					total_diffs++;
				}

				if ((dat1.flags & ROMOF_SUPPORT) && strcmp(games1[i].parent_name, games1[i].romof) &&
						strcmp(games1[i].romof, "neogeo") && strcmp(games1[i].romof, "playch10") && strcmp(games1[i].romof, "cvs") && strcmp(games1[i].romof, "decocass") && !(dat2.flags & ROMOF_SUPPORT))
				{
					/*printf("%s %s\n", games1[i].parent_name, games1[i].romof);*/
					bad_romof2++;
				}

				if ((dat2.flags & ROMOF_SUPPORT) && strcmp(games2[j].parent_name, games2[j].romof) &&
						strcmp(games2[j].romof, "neogeo") && strcmp(games2[j].romof, "playch10") && strcmp(games2[j].romof, "cvs") && strcmp(games2[j].romof, "decocass") && !(dat1.flags & ROMOF_SUPPORT))
				{
					/*printf("%s %s\n", games2[j].parent_name, games2[j].romof);*/
					bad_romof1++;
				}

				/* --- Load the ROMs from dat1 into buffer and sort --- */

				num_roms1=0;
				rom_ptr=games1[i].first_rom;
				while (rom_ptr && num_roms1<MAX_GAME_ROMS)
				{
					strcpy(roms1[num_roms1].name, rom_ptr->name);
					strcpy(roms1[num_roms1].merge, rom_ptr->merge);
					strcpy(roms1[num_roms1].size, rom_ptr->size);
					strcpy(roms1[num_roms1].crc, rom_ptr->crc);
					rom_ptr=rom_ptr->next_rom;
					num_roms1++;
				}
				if (rom_ptr)
					printf("Serious: %s has more ROMs than the limit (%d)\n", games1[i].name, MAX_GAME_ROMS);
				qsort(roms1, num_roms1, sizeof(struct rom), rom_sort_function);

				/* --- Load the ROMs from dat2 into buffer and sort --- */

				num_roms2=0;
				rom_ptr=games2[j].first_rom;
				while (rom_ptr && num_roms2<MAX_GAME_ROMS)
				{
					strcpy(roms2[num_roms2].name, rom_ptr->name);
					strcpy(roms2[num_roms2].merge, rom_ptr->merge);
					strcpy(roms2[num_roms2].size, rom_ptr->size);
					strcpy(roms2[num_roms2].crc, rom_ptr->crc);
					rom_ptr=rom_ptr->next_rom;
					num_roms2++;
				}
				if (rom_ptr)
					printf("Serious: %s has more ROMs than the limit (%d)\n", games2[j].name, MAX_GAME_ROMS);
				qsort(roms2, num_roms2, sizeof(struct rom), rom_sort_function);

				/* --- Compare the ROMs in buffers --- */

				k=l=0;

				while (k<num_roms1 || l<num_roms2 )
				{
					/* --- If reached the end of either, there are ROMs missing --- */

					if (k>=num_roms1) romdiff=1;
					if (l>=num_roms2) romdiff=-1;

					/* --- Else compare the ROM names --- */

					if (k<num_roms1 && l<num_roms2)
						romdiff=strcmp(roms1[k].name, roms2[l].name);

					/* --- If the names match, compare the ROM details --- */

					if (!romdiff)
					{
						/* --- Compare the merge names --- */

						if ((dat1.flags & MERGE_SUPPORT) && (dat2.flags & MERGE_SUPPORT) &&
							strcmp(roms1[k].merge, roms2[l].merge))
						{
							fprintf(out, "%s: %s has different merge names (%s - %s)\n", games1[i].name, roms1[k].name, roms1[k].merge, roms2[l].merge);
							total_diffs++;
						}

						if ((dat1.flags & MERGE_SUPPORT) && roms1[k].merge[0] && strcmp(roms1[k].name, roms1[k].merge) &&
								!(dat2.flags & MERGE_SUPPORT))
							bad_merge2++;

						if ((dat2.flags & MERGE_SUPPORT) && roms2[l].merge[0] && strcmp(roms2[l].name, roms2[l].merge) &&
								!(dat1.flags & MERGE_SUPPORT))
							bad_merge1++;

						/* --- Compare the sizes --- */

						if (roms1[k].size[0]!='\0' && roms2[l].size[0]!='\0' &&
							strcmp(roms1[k].size, roms2[l].size))
						{
							fprintf(out, "%s: %s has different sizes (%s - %s)\n", games1[i].name, roms1[k].name, roms1[k].size, roms2[l].size);
							total_diffs++;
						}

						if ((dat1.flags & dat2.flags & SIZE_SUPPORT) && roms1[k].size[0]=='\0')
							bad_size1++;

						if ((dat2.flags & dat2.flags & SIZE_SUPPORT) && roms2[l].size[0]=='\0')
							bad_size2++;

						/* --- Compare the CRCs --- */

						if ((dat1.flags & CRC_SUPPORT) && (dat2.flags & CRC_SUPPORT) &&
							strcmp(roms1[k].crc, roms2[l].crc))
						{
							fprintf(out, "%s: %s has different CRCs (%s - %s)\n", games1[i].name, roms1[k].name, roms1[k].crc, roms2[l].crc);
							total_diffs++;
						}

						/* --- Move on to next ROMs in buffers --- */

						k++; l++;
					}

					/* --- If ROM was missing in dat2 --- */

					if (romdiff<0)
					{
						if (dat1.flags & CRC_SUPPORT)
							fprintf(out, "%s: %s [%s] missing in %s\n", games1[i].name, roms1[k].name, roms1[k].crc, dat2.name);
						else
							fprintf(out, "%s: %s missing in %s\n", games1[i].name, roms1[k].name, dat2.name);
						k++; total_diffs++;
					}

					/* --- If ROM was missing in dat1 --- */

					if (romdiff>0)
					{
						if (dat2.flags & CRC_SUPPORT)
							fprintf(out, "%s: %s [%s] missing in %s\n", games2[j].name, roms2[l].name, roms2[l].crc, dat1.name);
						else
							fprintf(out, "%s: %s missing in %s\n", games2[j].name, roms2[l].name, dat1.name);
						l++; total_diffs++;
					}
				}

				/* --- Move on to next games --- */

				i++; j++;
			}

			/* --- If game was missing in dat2 --- */

			if (gamediff<0)
			{
				fprintf(out, "%s: **game** missing in %s\n", games1[i].name, dat2.name);
				i++; total_diffs++;
			}

			/* --- If game was missing in dat1 --- */

			if (gamediff>0)
			{
				fprintf(out, "%s: **game** missing in %s\n", games2[j].name, dat1.name);
				j++; total_diffs++;
			}
		}
	}

	/* --- If either dat uses alternative ROM names, use sort method 2 --- */

	if (!errflg && ((dat1.flags & ALT_ROM_SUPPORT) | (dat2.flags & ALT_ROM_SUPPORT)))
	{
		games1=dat1.games; games2=dat2.games;

		/* --- For all games --- */

		while (i<dat1.num_games || j<dat2.num_games)
		{
			/* --- If reached the end of either dat there is a game missing --- */

			if (i>=dat1.num_games) gamediff=1;
			if (j>=dat2.num_games) gamediff=-1;
			if (i<dat1.num_games && j<dat2.num_games)

			/* --- Else compare the game names --- */

				gamediff=strcmp(games1[i].name, games2[j].name);

			/* --- If the game names are the same then compare their details --- */

			if (gamediff==0)
			{
				/* --- Load the ROMs from dat1 into buffers  --- */
				/* --- One buffer for normal ROM details, one for alt names  --- */

				num_roms1=0;
				rom_ptr=games1[i].first_rom;
				alt_rom_ptr=games1[i].first_alt_rom;
				while (rom_ptr)
				{
					strcpy(roms1[num_roms1].name, rom_ptr->name);
					strcpy(roms1[num_roms1].size, rom_ptr->size);
					strcpy(roms1[num_roms1].crc, rom_ptr->crc);
					rom_ptr=rom_ptr->next_rom;

					/* --- If it has alt names, load them in buffer --- */

					if (dat1.flags & ALT_ROM_SUPPORT)
					{
						for (k=0; k<8; k++)
						{
							strcpy(alt_roms1[num_roms1].name[k], alt_rom_ptr->name[k]);
							strcpy(alt_roms1[num_roms1].size[k], alt_rom_ptr->size[k]);
						}
						alt_rom_ptr=alt_rom_ptr->next_alt_rom;
					}

					/* --- Else just store single name and clear alt names --- */

					else
					{
						strcpy(alt_roms1[num_roms1].name[0], roms1[num_roms1].name);
						strcpy(alt_roms1[num_roms1].size[0], roms1[num_roms1].size);
						for (k=1; k<8; k++)
						{
							alt_roms1[num_roms1].name[k][0]='\0';
							alt_roms1[num_roms1].size[k][0]='\0';
						}
					}

					num_roms1++;
				}

				/* --- Load the ROMs from dat2 into buffers  --- */
				/* --- One buffer for normal ROM details, one for alt names  --- */

				num_roms2=0;
				rom_ptr=games2[j].first_rom;
				alt_rom_ptr=games2[j].first_alt_rom;
				while (rom_ptr)
				{
					strcpy(roms2[num_roms2].name, rom_ptr->name);
					strcpy(roms2[num_roms2].size, rom_ptr->size);
					strcpy(roms2[num_roms2].crc, rom_ptr->crc);
					rom_ptr=rom_ptr->next_rom;

					/* --- If it has alt names, load them in buffer --- */

					if (dat2.flags & ALT_ROM_SUPPORT)
					{
						for (k=0; k<8; k++)
						{
							strcpy(alt_roms2[num_roms2].name[k], alt_rom_ptr->name[k]);
							strcpy(alt_roms2[num_roms2].size[k], alt_rom_ptr->size[k]);
						}
						alt_rom_ptr=alt_rom_ptr->next_alt_rom;
					}

					/* --- Else just store single name and clear alt names --- */

					else
					{
						strcpy(alt_roms2[num_roms2].name[0], roms2[num_roms2].name);
						strcpy(alt_roms2[num_roms2].size[0], roms2[num_roms2].size);
						for (k=1; k<8; k++)
						{
							alt_roms2[num_roms2].name[k][0]='\0';
							alt_roms2[num_roms2].size[k][0]='\0';
						}
					}

					num_roms2++;
				}

				/* --- For each ROM in buffer 1, check it is also in buffer 2 --- */

				for (k=0; k<num_roms1; k++)
				{
					found=0; sizediff=1;
					for (l=0; l<num_roms2; l++)
						for (m=0; alt_roms1[k].name[m][0]!='\0' && m<8; m++)
							for (n=0; alt_roms2[l].name[n][0]!='\0' && n<8; n++)
								if (!strcmp(alt_roms1[k].name[m], alt_roms2[l].name[n]))
								{
									found=1;
									if (alt_roms1[k].size[m][0]!='\0' && alt_roms2[l].size[n][0]!='\0')
										if (!strcmp(alt_roms1[k].size[m], alt_roms2[l].size[n]))
											sizediff=0;
										else
										{
											sprintf(size1, alt_roms1[k].size[m]);
											sprintf(size2, alt_roms2[l].size[n]);
										}
									else
										sizediff=0;

									if ((dat1.flags & dat2.flags & SIZE_SUPPORT) && roms1[k].size[0]=='\0')
										bad_size1++;

									if ((dat2.flags & dat2.flags & SIZE_SUPPORT) && roms2[l].size[0]=='\0')
										bad_size2++;
								}

					/* --- If not found, say so! --- */

					if (!found)
					{
						if (dat1.flags & CRC_SUPPORT)
							fprintf(out, "%s: %s [%s] missing in %s\n", games1[i].name, roms1[k].name, roms1[k].crc, dat2.name);
						else
							fprintf(out, "%s: %s missing in %s\n", games1[i].name, roms1[k].name, dat2.name);
						total_diffs++;
					}

					/* --- If size differed, say so! --- */

					if (found && sizediff)
					{
						fprintf(out, "%s: %s has different sizes (%s - %s)\n", games1[i].name, roms1[k].name, size1, size2);
						total_diffs++;
					}
				}
				
				/* --- For each ROM in buffer 2, check it is also in buffer 1 --- */

				for (k=0; k<num_roms2; k++)
				{
					found=0;
					for (l=0; l<num_roms1; l++)
						for (m=0; alt_roms2[k].name[m][0]!='\0' && m<8; m++)
							for (n=0; alt_roms1[l].name[n][0]!='\0' && n<8; n++)
								if (!strcmp(alt_roms2[k].name[m], alt_roms1[l].name[n]))
									found=1;

					/* --- If not found, say so! --- */

					if (!found)
					{
						if (dat2.flags & CRC_SUPPORT)
							fprintf(out, "%s: %s [%s] missing in %s\n", games1[i].name, roms2[k].name, roms2[k].crc, dat1.name);
						else
							fprintf(out, "%s: %s missing in %s\n", games1[i].name, roms2[k].name, dat1.name);
						total_diffs++;
					}

					/* --- No need to report differences, they were reported earlier --- */
				}
				
				/* --- Onto the next game in each dat --- */

				i++; j++;
			}

			/* --- If game was missing in dat2 --- */

			if (gamediff<0)
			{
				fprintf(out, "%s: **game** missing in %s\n", games1[i].name, dat2.name);
				i++; total_diffs++;
			}

			/* --- If game was missing in dat1 --- */

			if (gamediff>0)
			{
				fprintf(out, "%s: **game** missing in %s\n", games2[j].name, dat1.name);
				j++; total_diffs++;
			}
		}
	}

	/* --- Completion message --- */

	if (!errflg)
	{
		if (bad_romof1 || bad_romof2 || bad_size1 || bad_size2 || bad_merge1 || bad_merge2 ||
			dat1.num_samples || dat2.num_samples)
			fprintf(out, "\n");

		if (bad_romof1)	
		{
			fprintf(out, "Warning: %d games in %s use differing cloneof/romof names.\n", bad_romof1, dat2.name);
			printf("Warning: %d games in %s use differing cloneof/romof names.\n", bad_romof1, dat2.name);
		}

		if (bad_romof2)	
		{
			fprintf(out, "Warning: %d games in %s use differing cloneof/romof names.\n", bad_romof2, dat1.name);
			printf("Warning: %d games in %s use differing cloneof/romof names.\n", bad_romof2, dat1.name);
		}

		if (bad_size1)	
		{
			fprintf(out, "Warning: %d ROMs in %s don't specify their size.\n", bad_size1, dat1.name);
			printf("Warning: %d ROMs in %s don't specify their size.\n", bad_size1, dat1.name);
		}

		if (bad_size2)	
		{
			fprintf(out, "Warning: %d ROMs in %s don't specify their size.\n", bad_size2, dat2.name);
			printf("Warning: %d ROMs in %s don't specify their size.\n", bad_size2, dat2.name);
		}

		if (bad_merge1)	
		{
			fprintf(out, "Warning: %d ROMs in %s use differing merge names.\n", bad_merge1, dat2.name);
			printf("Warning: %d ROMs in %s use differing merge names.\n", bad_merge1, dat2.name);
		}

		if (bad_merge2)	
		{
			fprintf(out, "Warning: %d ROMs in %s use differing merge names.\n", bad_merge2, dat1.name);
			printf("Warning: %d ROMs in %s use differing merge names.\n", bad_merge2, dat1.name);
		}

		if (dat1.num_samples || dat2.num_samples)	
		{
			fprintf(out, "Warning: Sample comparison not supported.\n");
			printf("Warning: Sample comparison not supported.\n");
		}

		if (bad_romof1 || bad_romof2 || bad_size1 || bad_size2 || bad_merge1 || bad_merge2 ||
			dat1.num_samples || dat2.num_samples)
		{
			fprintf(out, "\n");
			printf("\n");
		}

		if (total_diffs)
		{
			fprintf(out, "Total of %d differences found.\n\n", total_diffs);
			printf("Comparison complete. %d differences reported in %s\n", total_diffs, useropts->outfile);
		}
		else
		{
			fprintf(out, "No differences found.\n\n");
			printf("Comparison complete. No differences found.\n");
		}

		/* --- Close log file --- */

		fclose(out);
	}

	return(0);
}

/* --------------------------------------------------------------------------
 * Sorting functions for qsort - not very complicated!
 * -------------------------------------------------------------------------- */

int game_sort_function(const void *game1, const void *game2)
{
	return(strcmp(game1, game2));
}

int parent_sort_function(const void *game1, const void *game2)
{
	int res;

	char *name1=((struct game *)(game1))->parent_name;
	char *name2=((struct game *)(game2))->parent_name;

	if (!*name1)
		name1=((struct game *)(game1))->name;

	if (!*name2)
		name2=((struct game *)(game2))->name;

	res=strcmp(name1, name2);

	if (!res)
	{
		if (!strcmp(game1, name1))
			res=-1;
		else if (!strcmp(game2, name2))
			res=1;
		else
			res=strcmp(game1, game2);
	}

	return(res);
}

int rom_sort_function(const void *rom1, const void *rom2)
{
	return(strcmp(rom1, rom2));
}

/* --------------------------------------------------------------------------
 * Dat initialization - Loads the first ten lines of the dat into a buffer
 * and calls each of the identification functions. If the dat is successfully
 * identified, the structure is then ready for the load routines.
 * -------------------------------------------------------------------------- */

int init_dat(struct dat *dat_info, char *fn)
{
	FILE *in;
	char lines[30][BUFFER_SIZE];
	int i=0, matches=0;
	struct stat buf;

	/* --- Initialize dat structure --- */

	strcpy(dat_info->name, fn);
	dat_info->num_games=dat_info->num_parents=dat_info->num_clones=0;
	dat_info->flags=dat_info->num_roms=dat_info->num_resource_roms=0;
	dat_info->num_samples=0;
	dat_info->games=0; dat_info->roms=0; dat_info->samples=0;

	/* --- Check if it is a directory --- */

	if (stat(fn, &buf) == 0)
	{
		if (buf.st_mode & S_IFDIR)
		{
			sprintf(dat_info->descr, "Directory Scan");
			dat_info->flags=LOAD_SUPPORT|SAVE_SUPPORT|SIZE_SUPPORT|CRC_SUPPORT;
			dat_info->load=dir_scan;
			dat_info->save=0;

			return(0);
		}
	}

	/* --- Open the dat file --- */

	if (!(in=fopen(dat_info->name, "r")))
	{
		printf("Error opening %s\n", dat_info->name);
		return(1);
	}

	/* --- Load the first ten lines --- */

	while (i<30 && fgets(lines[i], BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(lines[i]);
		i++;
	}

	/* --- Close the dat file --- */

	fclose(in);

	/* --- For all supported dat types, check if dat file matches --- */

	for (i=0;formats[i];i++)
	{
		if (formats[i]->identify && (formats[i]->identify)((char *)lines))
		{
			sprintf(dat_info->descr, formats[i]->descr);
			dat_info->flags=formats[i]->flags;
			dat_info->load=formats[i]->load;
			dat_info->save=formats[i]->save;
			matches++;
		}
	}

	/* --- If no matches, dat is not of supported format --- */

	if (matches==0)
		printf("Can't identify type for '%s'\n", dat_info->name);

	/* --- If >1 matches, blank flags to prevent further processing --- */

	if (matches>1)
	{
		printf("Can't identify type for '%s'. Matches more than one type.\n", dat_info->name);
		dat_info->flags=0;
	}

	return(0);
}

/* --------------------------------------------------------------------------
 * Function to identify parents of games in memory.
 * This is to make life easier later (when writing out a RomCenter dat).
 * Also identifies Neo-Geo games where this can only be seen from the parent.
 * -------------------------------------------------------------------------- */

int is_resource_rom(struct rom *curr_rom)
{
	/* Okay, I know that hard coded this is rubbish! */

	if (
		strcmp(curr_rom->crc, "9036d879") &&
		strcmp(curr_rom->crc, "97cf998b") &&
		strcmp(curr_rom->crc, "354029fc") &&
		/* Playchoice-10 */
		strcmp(curr_rom->crc, "d52fa07a") &&
		strcmp(curr_rom->crc, "30c15e23") &&
		strcmp(curr_rom->crc, "c1232eee") &&
		strcmp(curr_rom->crc, "9acffb30") &&
		strcmp(curr_rom->crc, "e5414ca3") &&
		strcmp(curr_rom->crc, "a2625c6e") &&
		strcmp(curr_rom->crc, "1213ebd4") &&
		/* CVS */
		strcmp(curr_rom->crc, "f055a624") &&
		strcmp(curr_rom->crc, "c205bca6") &&
		/* DECO Cassette */
		strcmp(curr_rom->crc, "a6df18fd") &&
		strcmp(curr_rom->crc, "238fdb40") &&
		strcmp(curr_rom->crc, "3b5836b4") &&
		strcmp(curr_rom->crc, "23d929b7") &&
		strcmp(curr_rom->crc, "b66b2c2a") &&
		strcmp(curr_rom->crc, "51eef657")
	)
		return(0);
	else
		return(1);
}

int identify_parents(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;
	struct game *games=dat_info->games;
	int num_games=dat_info->num_games;

	int num_parents=0;
	int unknown_parents=0;
	int i, j, parent_idx;
	struct rom *rom1, *rom2;
	struct rom *merge;
	int found;

	/* --- Unknown parents is allowable for formats such as System16 GCS --- */
	/* --- ...or there the user has specified an individual game --- */

	if (((dat_info->flags & ONE_GAME_ONLY) && games[0].parent_name[0]!='\0')||
		(num_games==1 && game &&
		!strcmp(games[0].name, game) && strcmp(games[0].name, games[0].parent_name) && games[0].parent_name[0]!='\0'))
	{
		dat_info->num_parents=0;
		printf("Warning: %s specifies an unknown parent (%s)\n\n",
			games[0].name, games[0].parent_name);
		return(0);
	}

	/* --- Otherwise, unknown parents is treated as a critical error --- */

	for (i=0;i<num_games; i++)
	{
		/* --- No point specifying this logic in every loader so it is here --- */

		if (!strcmp(games[i].name, "neogeo"))
			games[i].flags|= NEOGEO_GAME;

		/* --- If game specifies a parent --- */

		if (games[i].parent_name[0])
		{
			found=0;
 
			/* --- Search backward first (parent is likely to be before clone) --- */

			j=i-1;
			while (!found && j>=0)
			{
				if (!strcmp(games[j].name, games[i].parent_name))
				{
					parent_idx=j;
					strncpy(games[i].parent_title, games[j].title, TITLE_LENGTH);
					games[i].flags|=(games[j].flags & NEOGEO_GAME);

					/* --- If parent game not yet labelled as parent, do so --- */

					if (!(games[j].flags & PARENT_GAME))
					{
						games[j].flags|=PARENT_GAME;
						num_parents++;
					}
					found++;
				}

				j--;
			}

			/* --- Search forward if that didn't find it --- */

			j=i+1;
			while (!found && j<num_games)
			{
				if (!strcmp(games[j].name, games[i].parent_name))
				{
					parent_idx=j;
					strncpy(games[i].parent_title, games[j].title, TITLE_LENGTH);
					games[i].flags|=(games[j].flags & NEOGEO_GAME);

					/* --- If parent game not yet labelled as parent, do so --- */

					if (!(games[j].flags & PARENT_GAME))
					{
						games[j].flags|=PARENT_GAME;
						num_parents++;
					}
					found++;
				}

				j++;
			}

			if (!found)
			{
				/* --- IBM PC isn't declared in the MESS listinfo output--- */
 				if (!strcmp(games[i].parent_name, "pc"))
					sprintf(games[i].parent_title, "IBM PC");
				else
				{
					sprintf(games[i].parent_title, "Unknown");
					games[i].flags|=INVALID_PARENT;
				}

				rom1=games[i].first_rom;
				while (rom1 && ((useropts->options) & FIX_MERGING))
				{
					if (!is_resource_rom(rom1))
						rom1->merge[0]='\0';
					rom1=rom1->next_rom;
				}
			}
			else
			{
				rom1=games[i].first_rom;
				while (rom1)
				{
					merge=0; found=0;
					rom2=games[parent_idx].first_rom;
					while (rom2)
					{
						if (strcmp(rom1->crc, "00000000") &&
							!strcmp(rom1->crc, rom2->crc))
						{
							if (!found && strcmp(rom1->merge, rom2->name))
								merge=rom2;
							if (!strcmp(rom1->merge, rom2->name))
								found++;
						}
						rom2=rom2->next_rom;
					}
					if (!found && merge)
					{
						/*printf("%s,%s,%s\n", games[i].name, rom1->name, merge->name);
						strcpy(rom1->merge, merge->name);*/
					}

					merge=0; found=0;
					rom2=games[parent_idx].first_rom;
					if ((useropts->options) & FIX_MERGING && !is_resource_rom(rom1))
						rom1->merge[0]='\0';
					while (((useropts->options) & FIX_MERGING) && rom2)
					{
						if (!strcmp(rom1->name, rom2->name))
						{
							if (strcmp(rom1->crc, rom2->crc) &&
								(~strtoul(rom1->crc, NULL, 16) == strtoul(rom2->crc, NULL, 16) ||
								!strcmp("00000000", rom2->crc) || !strcmp(rom1->crc, "00000000")))
							{
								//printf("CRC conflict for %s (%s in %s - %s in %s)\n", rom2->name, rom2->crc, games[parent_idx].name, rom1->crc, games[i].name);
								if (!strcmp("00000000", rom2->crc))
									strcpy(rom2->crc, rom1->crc);
								else
									strcpy(rom1->crc, rom2->crc);
							}

							strcpy(rom1->merge, rom1->name);
						}
						rom2=rom2->next_rom;
					}

					rom1=rom1->next_rom;
				}
			}
		}
		else
		{
			rom1=games[i].first_rom;
			while (rom1 && ((useropts->options) & FIX_MERGING))
			{
				if (!is_resource_rom(rom1))
					rom1->merge[0]='\0';
				rom1=rom1->next_rom;
			}
		}
	}

	/* --- Update details in dat header for number of parents --- */

	dat_info->num_parents=num_parents;

	/* --- Return > 0 indicates a fatal error --- */

	return(unknown_parents);
}

/* --------------------------------------------------------------------------
 * Function to display summary information about the dat held in memory
 * -------------------------------------------------------------------------- */

int display_dat_info(struct dat *dat_info)
{
	printf("Name:      %s\n", dat_info->name);
	printf("Type:      %s\n\n", dat_info->descr);

	printf("Games:     %d (%d parents, %d clones, %d other)\n", dat_info->num_games-dat_info->num_resources, dat_info->num_parents, dat_info->num_clones, dat_info->num_games-dat_info->num_parents-dat_info->num_clones-dat_info->num_resources);
	printf("Resources: %d\n", dat_info->num_resources);
	printf("ROMs:      %ld (%ld for resources)\n", dat_info->num_roms, dat_info->num_resource_roms);
	printf("Samples:   %d\n", dat_info->num_samples);
	printf("-------------------------------------------------------------------------------\n");

	return(0);
}

/* --------------------------------------------------------------------------
 * Function to release dat from memory
 * -------------------------------------------------------------------------- */

int release_dat(struct dat *dat_info)
{
	if (dat_info->samples) free(dat_info->samples);
	if (dat_info->alt_roms) free(dat_info->alt_roms);
	if (dat_info->roms) free(dat_info->roms);
	if (dat_info->games) free(dat_info->games);

	return(0);
}

/* --------------------------------------------------------------------------
 * MAME Listinfo - Load and Save
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_mame_listinfo(char *lines)
{
	int i;

	/* --- If line exists with 'game (' or 'resource (', not 'emulator (' --- */

	for (i=0; i<20; i++)
	{
		if (!strncmp(lines+i*BUFFER_SIZE, "game (", 6))
			return(1);
		if (!strncmp(lines+i*BUFFER_SIZE, "resource (", 10))
			return(1);
		if (!strncmp(lines+i*BUFFER_SIZE, "emulator (", 10))
			return(0);
	}

	return(0);
}

/* --- Load --- */

int load_mame_listinfo(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;
	int clones=(useropts->options) & GAME_AND_CLONES;
	int remove=(useropts->options) & REMOVE_CLONES;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		SKIP_WHITESPACE(st, st_ptr)                /* Some dats have indentation */

		if (!strncmp(st_ptr, "resource", 8))
			HOLD_GAME_FLAG(RESOURCE)                 /* Neo-Geo ROMs */

		if (!strncmp(st_ptr, "machine", 7))
			HOLD_GAME_FLAG(MACHINE)                  /* Used by MESS */

		if (!strncmp(st_ptr, "clrmamepro", 10))
			HOLD_GAME_FLAG(CLRMAMEPRO)               /* Used by ClrMamePro */

		if ((flags_hold & CLRMAMEPRO) && !strncmp(st_ptr, "category", 8))
		{
			GET_TOKEN(token, st_ptr)                 /* skip category token */
			GET_TOKEN(token, st_ptr)
			strcpy(comment, token);
		}

		if ((flags_hold & CLRMAMEPRO) && !strncmp(st_ptr, "version", 7))
		{
			GET_TOKEN(token, st_ptr)                 /* skip version token */
			GET_TOKEN(token, st_ptr)
			strcpy(version, token);
		}

		if ((flags_hold & CLRMAMEPRO) && !strncmp(st_ptr, "author", 6))
		{
			GET_TOKEN(token, st_ptr)                 /* skip author token */
			GET_TOKEN(token, st_ptr)
			strcpy(author, token);
		}

		if ((flags_hold & CLRMAMEPRO) && !strncmp(st_ptr, "forcemerging", 12))
		{
			GET_TOKEN(token, st_ptr)                 /* skip force token */
			GET_TOKEN(token, st_ptr)
			strcpy(merging, token);
		}

		if ((flags_hold & CLRMAMEPRO) && !strncmp(st_ptr, "forcezipping", 12))
		{
			GET_TOKEN(token, st_ptr)                 /* skip force token */
			GET_TOKEN(token, st_ptr)
			strcpy(zipping, token);
		}

		if (!strncmp(st_ptr, "name", 4))
		{
			GET_TOKEN(token, st_ptr)                 /* skip name token */
			GET_TOKEN(token, st_ptr)
			HOLD_GAME_NAME(token)                    /* Remember the game name */
		}

		if (!strncmp(st_ptr, "description", 11))
		{
			GET_TOKEN(token, st_ptr)                 /* skip description token */
			GET_TOKEN(token, st_ptr)
			HOLD_GAME_TITLE(token)                   /* Hold description */
		}

		if (!strncmp(st_ptr, "year", 4))
		{
			GET_TOKEN(token, st_ptr)                 /* skip year token */
			GET_TOKEN(token, st_ptr)
			HOLD_YEAR(token)                         /* Hold year */
		}

		if (!strncmp(st_ptr, "manufacturer", 12))
		{
			GET_TOKEN(token, st_ptr)                 /* skip manufacturer token */
			GET_TOKEN(token, st_ptr)
			HOLD_MANUFACTURER(token)                 /* Hold manufacturer */
		}

		if (!strncmp(st_ptr, "cloneof", 7))
		{
			GET_TOKEN(token, st_ptr)                 /* skip cloneof token */
			GET_TOKEN(token, st_ptr)
			HOLD_PARENT_NAME(token)                  /* ...and parent name */
		}

		if (!strncmp(st_ptr, "romof", 5))
		{
			GET_TOKEN(token, st_ptr)                 /* skip romof token */
			GET_TOKEN(token, st_ptr)
			HOLD_ROMOF(token)                        /* ...and romof */
		}

		if (!strncmp(st_ptr, "rom (", 5))
			STORE_ROM(st_ptr)                        /* Only stores if interested! */

		if (!strncmp(st_ptr, "sampleof", 8))
		{
			GET_TOKEN(token, st_ptr)                 /* skip sampleof token */
			GET_TOKEN(token, st_ptr)
			HOLD_SAMPLEOF(token)                     /* ...and sampleof */
		}

		if (!strncmp(st_ptr, "sample ", 7))
			STORE_SAMPLE(st_ptr+7)                   /* Only stores if interested! */

		if (st_ptr[0]==')')                        /* No more info for game */
			STORE_GAME                               /* Only stores if interested! */
	}

	DONE_LOAD
}

/* --- Save --- */

int save_mame_listinfo(struct dat *dat_info, struct options *useropts)
{
	struct game *games=dat_info->games;
	FILE *out, *log;
	int i, j, bits;
	long size;
	int funny_size=0, bad_size=0, bad_parent=0, warning=1;
	struct rom *curr_rom;
	struct sample *curr_sample;

	/* --- Open file for writing --- */

	if (!(out=fopen(useropts->outfile, useropts->outmode)))
	{
		printf("Error opening %s\n", useropts->outfile);
		return(1);
	}

	if (!(log=fopen("datutil.log", useropts->outmode)))
	{
		fclose(out);
		printf("Error opening %s\n", "datutil.log");
		return(1);
	}

	/* --- Tell user what is going on --- */

	if (useropts->outmode[0]=='w')
		printf("Writing");
	else
		printf("Appending to");

	printf(" %s in MAME Listinfo format...\n", useropts->outfile);

	/* --- For each game... --- */

	if (useropts->rc_refname || useropts->rc_fullname || useropts->rc_zipping || useropts->rc_merging)
	{
		fprintf(out, "clrmamepro (\n");

		if (useropts->rc_refname)
		{
			if (!strchr(useropts->rc_refname, ' '))
				fprintf(out, "\tname %s\n", useropts->rc_refname);
			else
				fprintf(out, "\tname \"%s\"\n", useropts->rc_refname);
		}

		if (useropts->rc_fullname)
			fprintf(out, "\tdescription \"%s\"\n", useropts->rc_fullname);

		if (useropts->rc_comment)
		{
			if (!strchr(useropts->rc_comment, ' '))
				fprintf(out, "\tcategory %s\n", useropts->rc_comment);
			else
				fprintf(out, "\tcategory \"%s\"\n", useropts->rc_comment);
		}

		if (useropts->rc_version)
		{
			if (!strchr(useropts->rc_version, ' '))
				fprintf(out, "\tversion %s\n", useropts->rc_version);
			else
				fprintf(out, "\tversion \"%s\"\n", useropts->rc_version);
		}

		if (useropts->rc_author)
		{
			if (!strchr(useropts->rc_author, ' '))
				fprintf(out, "\tauthor %s\n", useropts->rc_author);
			else
				fprintf(out, "\tauthor \"%s\"\n", useropts->rc_author);
		}

		if (useropts->rc_merging)
			fprintf(out, "\tforcemerging %s\n", useropts->rc_merging);

		if (useropts->rc_zipping)
			fprintf(out, "\tforcezipping %s\n", useropts->rc_zipping);

		fprintf(out, ")\n\n");
	}

	for (i=0; i<dat_info->num_games; i++)
	{
		/* --- Check for invalid parent --- */
		if (games[i].flags & INVALID_PARENT)
		{
			fprintf(log, "%s: specifies an invalid parent ('%s')\n", games[i].name, games[i].parent_name);
			bad_parent++;
		}

		/* --- It may be a game or resource (neogeo) --- */

		if (games[i].flags & RESOURCE)
			fprintf(out, "resource (\n");
		/*else if (games[i].flags & MACHINE) - Not ClrMame compatible
			fprintf(out, "machine (\n");*/
		else
			fprintf(out, "game (\n");

		/* --- Write its name, description, parent etc --- */

		if (strchr(games[i].name, ' '))
			fprintf(out, "\tname \"%s\"\n", games[i].name);
		else
			fprintf(out, "\tname %s\n", games[i].name);
		fprintf(out, "\tdescription \"%s\"\n", games[i].title);
		if (games[i].year[0])
			fprintf(out, "\tyear %s\n", games[i].year);
		if (games[i].manufacturer[0])
			fprintf(out, "\tmanufacturer \"%s\"\n", games[i].manufacturer);
		if (games[i].parent_name[0])
		{
			if (strchr(games[i].parent_name, ' '))
				fprintf(out, "\tcloneof \"%s\"\n", games[i].parent_name);
			else
				fprintf(out, "\tcloneof %s\n", games[i].parent_name);
		}
		if (games[i].romof[0])
		{
			if (strchr(games[i].romof, ' '))
				fprintf(out, "\tromof \"%s\"\n", games[i].romof);
			else
				fprintf(out, "\tromof %s\n", games[i].romof);
		}

		/* --- Write out its ROM details --- */

		if ((curr_rom=games[i].first_rom))
		{
			while (curr_rom)
			{
				if (strchr(curr_rom->name, ' '))
					fprintf(out, "\trom ( name \"%s\" ", curr_rom->name);
				else
					fprintf(out, "\trom ( name %s ", curr_rom->name);

				/* --- Merged ROMs have an extra bit --- */

				if (curr_rom->merge[0])
				{
					if (strchr(curr_rom->merge, ' '))
						fprintf(out, "merge \"%s\" ", curr_rom->merge);
					else
						fprintf(out, "merge %s ", curr_rom->merge);
				}

				if (curr_rom->size[0])
				{
					size=atoi(curr_rom->size); bits=0;
					for (j=0; j < 32; j++)
					{
						if ((1<<j) & size)
							bits++;
					}
					if (bits!=1)
					{
						fprintf(log, "%s: %s has a funny size ('%s')\n", games[i].name, curr_rom->name, curr_rom->size);
						funny_size++;
					}
					fprintf(out, "size %s ", curr_rom->size);
				}
				else
				{
					fprintf(log, "%s: %s has no size\n", games[i].name, curr_rom->name);
					bad_size++;
				}

				fprintf(out, "crc %s )\n", curr_rom->crc);

				curr_rom=curr_rom->next_rom;
			}
		}

		/* --- Write out its Sample details --- */

		if (games[i].sampleof[0])
			fprintf(out, "\tsampleof %s\n", games[i].sampleof);

		if ((curr_sample=games[i].first_sample))
		{
			while (curr_sample)
			{
				fprintf(out, "\tsample %s\n",curr_sample->name);
				curr_sample=curr_sample->next_sample;
			}
		}
		
		/* --- End of details for the game --- */

		fprintf(out, ")\n\n");
	}

	/* --- Close file --- */

	fclose(log);
	fclose(out);

	/* --- Display summary to user --- */

	if (bad_parent || funny_size || bad_size)
	{
		printf("\nWarnings (see datutil.log):\n\n");

		if (bad_parent)	
		{
			printf("%d) %d games have an invalid parent.\n", warning, bad_parent);
			warning++;
		}

		if (funny_size)	
		{
			printf("%d) %d ROMs have a funny size.\n", warning, funny_size);
			if (funny_size==dat_info->num_roms)
				printf("   Try changing the RomCenter DAT version to 1.92 (use a text editor)!\n");
			warning++;
		}

		if (bad_size)	
		{
			printf("%d) %d ROMs don't specify their size.\n", warning, bad_size);
			warning++;
		}
	}

	printf("\nAll done!\n");

	return(0);
}

/* --------------------------------------------------------------------------
 * Title List
 * -------------------------------------------------------------------------- */

/* --- Save --- */

int save_title_list(struct dat *dat_info, struct options *useropts)
{
	struct game *games=dat_info->games;
	FILE *out, *log;
	int bad_parent=0, warning=1;
	int i, j;

	/* --- Open file for writing --- */

	if (!(out=fopen(useropts->outfile, useropts->outmode)))
	{
		printf("Error opening %s\n", useropts->outfile);
		return(1);
	}

	if (!(log=fopen("datutil.log", useropts->outmode)))
	{
		fclose(out);
		printf("Error opening %s\n", "datutil.log");
		return(1);
	}

	/* --- Tell user what is going on --- */

	if (useropts->outmode[0]=='w')
		printf("Writing");
	else
		printf("Appending to");

	printf(" %s as a title list...\n", useropts->outfile);

	/* --- For each game... --- */

	for (i=0; i<dat_info->num_games; i++)
	{
		/* --- Check for invalid parent --- */
		if (games[i].flags & INVALID_PARENT)
		{
			fprintf(log, "%s: specifies an invalid parent ('%s')\n", games[i].name, games[i].parent_name);
			bad_parent++;
		}

		/* --- Write its name, description, parent etc --- */

		if (games[i].parent_name[0])
			fprintf(out, "  ");
		fprintf(out, "%s ", games[i].name);

		if (strlen(games[i].name)<10)
			for (j=0; j<10-strlen(games[i].name); j++)
				fprintf(out, " ");
		fprintf(out, "%s\n", games[i].title);

		/* --- End of details for the game --- */
	}

	/* --- Close file --- */

	fclose(log);
	fclose(out);

	/* --- Display summary to user --- */

	if (bad_parent)
	{
		printf("\nWarnings (see datutil.log):\n\n");

		if (bad_parent)	
		{
			printf("%d) %d games have an invalid parent.\n", warning, bad_parent);
			warning++;
		}
	}

	printf("\nAll done!\n");

	return(0);
}

/* --------------------------------------------------------------------------
 * MESS Listinfo - Load only (use MAME Listinfo for saving)
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_mess_listinfo(char *lines)
{
	int i;

	/* --- If first line begins with 'machine (' --- */

	for (i=0; i<20; i++)
		if (!strncmp(lines+i*BUFFER_SIZE, "machine (", 9))
			return(1);

	return(0);
}

/* --- Load (use MAME Listinfo loader) --- */

int load_mess_listinfo(struct dat *dat_info, struct options *useropts)
{
	return (load_mame_listinfo(dat_info, useropts));
}

/* --- Save (use MAME Listinfo saver) --- */

int save_mess_listinfo(struct dat *dat_info, struct options *useropts)
{
	return(save_mame_listinfo(dat_info, useropts));
}

/* --------------------------------------------------------------------------
 * Shark Listinfo
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_shark_listinfo(char *lines)
{
	/* --- If first line begins with 'emulator ( "Shark' --- */

	if (!strncmp(lines, "emulator ( \"Shark", 16))
		return(1);

	return(0);
}

/* --- Load --- */

int load_shark_listinfo(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		SKIP_WHITESPACE(st, st_ptr)                /* Some dats have indentation */

		if (!strncmp(st_ptr, "name", 4))
			HOLD_GAME_NAME(st_ptr+5)                 /* Remember the game name */

		if (!strncmp(st_ptr, "description", 11))
		{
			REMOVE_QUOTES(st_ptr+12)                 /* Remove quotes (if required) */
			HOLD_GAME_TITLE(st_ptr+12)               /* Hold description */
		}

		if (!strncmp(st_ptr, "rom (", 5))
		{
			LOWER(st_ptr)
			STORE_ROM(st_ptr)                        /* Only stores if interested! */
		}

		if (st_ptr[0]==')')                        /* No more info for this game */
			STORE_GAME  				                     /* Only stores if interested! */
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * RAINE Gameinfo
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_raine_gameinfo(char *lines)
{
	int i;

	/* --- if contains 'name "Raine' --- */

	for (i=0; i<20; i++)
		if (strstr(lines+i*BUFFER_SIZE, "name \"Raine"))
			return(1);

	return(0);
}

/* --- Load --- */

int load_raine_gameinfo(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;
	int clones=(useropts->options) & GAME_AND_CLONES;
	int remove=(useropts->options) & REMOVE_CLONES;

	int emulator=0;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		SKIP_WHITESPACE(st, st_ptr)                /* Some dats have indentation */

		if (!strncmp(st_ptr, "emulator (", 10))
			emulator=1;                              /* We aren't interested */

		if (!strncmp(st_ptr, "name", 4))
		{
			GET_TOKEN(token, st_ptr)                 /* skip name token */
			GET_TOKEN(token, st_ptr)
			HOLD_GAME_NAME(token)                    /* Remember the game name */
		}

		if (!strncmp(st_ptr, "description", 11))
		{
			if (strstr(st_ptr, " japanese \""))
				*strstr(st_ptr, " japanese \"")='\0';  /* Skip Japenese version */

			GET_TOKEN(token, st_ptr)                 /* skip description token */
			GET_TOKEN(token, st_ptr)
			HOLD_GAME_TITLE(token)                   /* Hold description */
		}

		if (!strncmp(st_ptr, "year", 4))
		{
			GET_TOKEN(token, st_ptr)                 /* skip year token */
			GET_TOKEN(token, st_ptr)
			HOLD_YEAR(token)                         /* Hold year */
		}

		if (!strncmp(st_ptr, "manufacturer", 12))
		{
			GET_TOKEN(token, st_ptr)                 /* skip manufacturer token */
			GET_TOKEN(token, st_ptr)
			HOLD_MANUFACTURER(token)                 /* Hold manufacturer */
		}

		if (!strncmp(st_ptr, "cloneof", 7))
		{
			GET_TOKEN(token, st_ptr)                 /* skip cloneof token */
			GET_TOKEN(token, st_ptr)
			HOLD_PARENT_NAME(token)                  /* ...and parent name */
		}

		if (!strncmp(st_ptr, "romof", 5))
		{
			GET_TOKEN(token, st_ptr)                 /* skip romof token */
			GET_TOKEN(token, st_ptr)
			HOLD_ROMOF(token)                        /* ...and romof */
		}

		if (!strncmp(st_ptr, "rom (", 5))
			STORE_ROM(st_ptr)                        /* Only stores if interested! */

		if (st_ptr[0]==')')                        /* No more info for this game */
		{
			if (!emulator)
				STORE_GAME			                       /* Only stores if interested! */
			else
				emulator=0;                            /* We are past the emulator */
		}
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * Callus Gamelist
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_callus_gamelist(char *lines)
{
	/* --- If second line contains 'Callus' --- */

	if (!strstr(lines, "clrmamepro") && strstr(lines+BUFFER_SIZE, "Callus"))
		return(1);

	return(0);
}

/* --- Load --- */

int load_callus_gamelist(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;
	int clones=(useropts->options) & GAME_AND_CLONES;
	int remove=(useropts->options) & REMOVE_CLONES;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		if (st[0]==' ')                          /* Game title preceded by spaces */
			HOLD_GAME_TITLE(st)                    /* ...remember it! */

		if (!strncmp(st, "Name:", 5))
			HOLD_GAME_NAME(st+14)                  /* ...and the game name */

		if (!strncmp(st, "Directory:", 10) ||
		    !strncmp(st, "Parent:", 7))
			HOLD_PARENT_NAME(st+14)                /* ...and the parent name */

		if (st[0]=='	')                         /* ROM names are preceded by TAB */
		{
			LOWER(st)
			STORE_ROM(st+1)                        /* Only stores if interested! */
		}

		if (st[0]=='-')                          /* No more info for this game */
			STORE_GAME                             /* Only stores if interested! */
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * JFF Text
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_jff_txt(char *lines)
{
	int i;

	/* --- If any of the first ten lines contain 'JUST FOR FUN' --- */

	for (i=0; i<10; i++)
		if (strstr(lines+BUFFER_SIZE*i, "JUST FOR FUN"))
			return(1);

	return(0);
}

/* --- Load --- */

int load_jff_txt(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;

	int rom_flag=0;
	char rom_string[80];
	char rom_name[FILE_NAME_LENGTH+1];
	char rom_size[20];

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		if (!strncmp(st, "###", 3))              /* Game title preceded by hashes */
		{
			st[strlen(st)-4]='\0';
			HOLD_GAME_TITLE(st+4)                  /* ...remember it! */
		}

		if (!strcmp(st+4, "directory name:"))
		{
			fgets(st, BUFFER_SIZE, in);
			fgets(st, BUFFER_SIZE, in);
			REMOVE_CR_LF(st)
			LOWER(st)
			HOLD_GAME_NAME(st)                     /* ...and the game name */
		}

		if (!strcmp(st+4, "names:"))
		{
			fgets(st, BUFFER_SIZE, in);
			REMOVE_CR_LF(st)
			while (st[0]=='\0')
			{
				fgets(st, BUFFER_SIZE, in);
				REMOVE_CR_LF(st)
			}
			rom_flag=1;                            /* next few lines are ROMs */
		}

		if (rom_flag && st[0]!='\0')             /* This is a ROM */
		{
			if (strchr(st, ';'))
				*strchr(st, ';')='\0';
			LOWER(st)
			EXTRACT_NAME(rom_name, st)
			EXTRACT_SIZE(rom_size, st+14)
			sprintf(rom_string, "rom ( name %s size %s )", rom_name, rom_size);
			STORE_ROM(rom_string)                  /* Only stores if interested! */
		}

		if (rom_flag && st[0]=='\0')             /* No more info for this game */
		{	
			STORE_GAME                             /* Only stores if interested! */
			rom_flag=0;
		}
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * Sparcade Romlist - Not enabled as romlist.txt is not 100% correct.
 * It doesn't specify extensions correctly for some games, some game names are
 * incorrect and it misses some ROMs. These are all correct in the .RAT files.
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_sparcade_romlist(char *lines)
{
	int i;

	/* --- If any of the first ten lines contain 'Sparcade' --- */

	for (i=0; i<10; i++)
		if (strstr(lines+BUFFER_SIZE*i, "Sparcade"))
			return(1);

	return(0);
}

/* --- Load --- */

int load_sparcade_romlist(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;

	int past_example=0;
	char rom_string[120];

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		if (!strcmp(st, "-----"))                 /* Past the example */
			past_example=1;

		if (past_example && strchr(st, '['))
		{
			if (strchr(st, ']'))
				*strrchr(st, ']')='\0';
			HOLD_GAME_NAME(strchr(st, '[')+1)      /* hold the game name */
			*strrchr(st, ' ')='\0';
			HOLD_GAME_TITLE(st);                   /* ...and the game title */
			fgets(st, BUFFER_SIZE, in);            /* skip the next line */
			fgets(st, BUFFER_SIZE, in);            /* get the next line */
			REMOVE_CR_LF(st)
		}

		if (past_example && !strncmp(st, "  ", 2))  /* This is a ROM */
		{
			LOWER(st)
			EXTRACT_NAMES(rom_string, st, "   ")   /* Ensure there is only one space between names */
			STORE_ROM(rom_string)                  /* Only stores if interested! */
		}

		if (past_example && st[0]=='\0')         /* No more info for this game */
			STORE_GAME                             /* Only stores if interested! */
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * Sparcade RAT File
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_sparcade_rat(char *lines)
{
	int i;

	/* --- If any of the first ten lines begin 'path:' --- */

	for (i=0; i<10; i++)
		if (!strncmp(lines+BUFFER_SIZE*i, "path:", 5))
			return(1);

	return(0);
}

/* --- Load --- */

int load_sparcade_rat(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;

	char *name_ptr;
	char rom_string[120];
	char ext[10]="";

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		RTRIM(st)

		if (!strncmp(st, "path:", 5))
		{
	    STORE_GAME                             /* Only stores if interested! */
			ext[0]='\0';
			name_ptr=st+5;
			while (name_ptr[0]==' ' || name_ptr[0]=='	')
				name_ptr++;
			HOLD_GAME_NAME(name_ptr)               /* hold the game name */
		}

		if (!strncmp(st, "ext:", 4))
		{
			name_ptr=st+5;
			while (name_ptr[0]==' ' || name_ptr[0]=='	')
				name_ptr++;
			strncpy(ext, name_ptr, 3);             /* hold the game ext */
			ext[4]='\0';
		}

		if (st[0] && st[0]!=' ' && st[0]!='	' &&
        st[0]!='#' && (!strchr(st, ':')))    /* This is a ROM */
		{
			LOWER(st)
			EXTRACT_NAMES(rom_string, st, ext)     /* Ensure there is only one space between names */
			STORE_ROM(rom_string)                  /* Only stores if interested! */
		}
	}

	STORE_GAME                                 /* Only stores if interested! */

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * JAS-SYS1 Roms Text
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_jas_romlist(char *lines)
{
	/* --- Check text on the third line --- */

	if (!strncmp(lines+BUFFER_SIZE*2, "Directory", 9))
		return(1);

	return(0);
}

/* --- Load --- */

int load_jas_romlist(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;

	char *name_ptr;
	int rom_flag=0;
	char rom_string[80];
	char rom_name[FILE_NAME_LENGTH+1];
	char rom_size[20];

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		RTRIM(st)                                 /* There are some lines with just spaces */

		if (title_hold[0]=='\0' && st[0]!='\0')
			HOLD_GAME_TITLE(st)                     /* ...remember it! */

		if (!strncmp(st, "Directory", 9))
		{	
			LOWER(st)
			name_ptr=st+10;
			while (name_ptr[0]==' ' || name_ptr[0]=='	')
				name_ptr++;
			HOLD_GAME_NAME(name_ptr)               /* ...and the game name */
			fgets(st, BUFFER_SIZE, in);            /* skip the next line */
			fgets(st, BUFFER_SIZE, in);            /* skip the next line */
			fgets(st, BUFFER_SIZE, in);            /* get the next line */
			REMOVE_CR_LF(st)
			rom_flag=1;                            /* Okay, start looking for ROMs now! */
		}

		if (rom_flag && st[0]!='\0')
		{                                        /* This is a ROM */
			LOWER(st)
			EXTRACT_NAME(rom_name, st)
			EXTRACT_SIZE(rom_size, st+14)
			sprintf(rom_string, "rom ( name %s size %s )", rom_name, rom_size);
			STORE_ROM(rom_string)                  /* Only stores if interested! */
		}

		if (st[0]=='\0')                         /* No more info for this game */
		{	
				STORE_GAME                           /* Only stores if interested! */
				HOLD_GAME_TITLE("")
				rom_flag=0;
		}
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * Retrocade Romlist
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_retrocade_romlist(char *lines)
{
	/* --- If first line begins 'Game:' --- */

	if (!strncmp(lines, "Game:", 5))
		return(1);

	return(0);
}

/* --- Load --- */

int load_retrocade_romlist(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;
	struct rom *rom_ptr;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		if (!strncmp(st, "Game:", 5))
		{
			*strrchr(st, '.')='\0';        /* Remove '.zip )' from line */
			LOWER(st+58)
			HOLD_GAME_NAME(st+58)          /* Game name is now at end of line */
			*strrchr(st, '(')='\0';        /* Remove bracket infront of it */
			RTRIM(st)                      /* ...and the spaces before */
			HOLD_GAME_TITLE(st+6)          /* Game title is now at right of line */
		}

		if (!strncmp(st, "  ", 2))
		{
			*strchr(st, '(')='\0';         /* Remove status of rom */
			RTRIM(st)                      /* ...and spaces */
			LOWER(st)
			STORE_ROM_UNIQUE(st+8)         /* Only stores if interested! */
		}

		if (st[0]=='\0')                 /* Empty line signifies end of game */
			STORE_GAME                     /* Only stores if interested! */
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * System16 GCS - Note that the shortname is incorrect in SHO_BOOT.GCS
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_system16_gcs(char *lines)
{
	int i;

	/* --- If any of the first ten lines begins with '.game' --- */

	for (i=0; i<10; i++)
		if (!strncmp(lines+i*BUFFER_SIZE, ".game", 5))
			return(1);

	return(0);
}

/* --- Load --- */

int load_system16_gcs(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;
	int clones=(useropts->options) & GAME_AND_CLONES;
	int remove=(useropts->options) & REMOVE_CLONES;
	struct rom *rom_ptr;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		if (!strncmp(st, ".game", 5))
			HOLD_GAME_TITLE(st+6)                  /* Remember the title */

		if (!strncmp(st, ".shortname", 10))
		{
			LOWER(st+11)
			HOLD_GAME_NAME(st+11)                  /* ...and the game name */
		}

		if (!strncmp(st, ".zip", 4))
		{
			*strrchr(st, '.')='\0';                /* Remove '.zip' */
			HOLD_PARENT_NAME(st+5)                 /* Remember the parent name */
		}

		if (!strncmp(st, ".romload", 8) ||
				!strncmp(st, ".scrload", 8) ||
				!strncmp(st, ".objload", 8) ||
				!strncmp(st, ".objload2", 9) ||
				!strncmp(st, ".z80load", 8))
			STORE_S16_ROMS(st, 1, 99)              /* Only stores if interested! */

		if (!strncmp(st, ".z80data", 8))
			STORE_S16_ROMS(st, 2, 99)              /* Only stores if interested! */

		if (!strncmp(st, ".loaddata", 9) ||
			!strncmp(st, ".speech", 7))
			STORE_S16_ROMS(st, 3, 99)              /* Only stores if interested! */


		if (!strncmp(st, ".grscrload", 10))
			STORE_S16_ROMS(st, 1, 1)               /* Only stores if interested! */

		if (!strncmp(st, ".end", 4))             /* No more info for this game */
			STORE_GAME                             /* Only stores if interested! */
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * Calice Driver
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_calice_drv(char *lines)
{
	int i;

	/* --- If any of the first thirty lines contains 'GAMENAME' --- */

	for (i=0; i<30; i++)
		if (!strncmp(lines+i*BUFFER_SIZE, "GAMENAME", 8))
			return(1);

	return(0);
}

/* --- Load --- */

int load_calice_drv(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;
	int clones=(useropts->options) & GAME_AND_CLONES;
	int remove=(useropts->options) & REMOVE_CLONES;
	struct rom *rom_ptr;
	int in_z80, in_samples;
	int found_games=0;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		SKIP_WHITESPACE(st, st_ptr)                /* Some dats have indentation */

		if (st[0]=='[')
		{
			in_samples=in_z80=0;
			LOWER(st)
		}

		if (!strncmp(st, "[z80]", 5))
			in_z80=1;

		if (!strncmp(st, "[samples]", 9))
			in_samples=1;

		if (!strncmp(st, "GAMENAME", 8))
		{
			GET_TOKEN(token, st_ptr)                /* skip name token */
			GET_TOKEN(token, st_ptr)                /* skip name token */
			HOLD_GAME_TITLE(token)                  /* Remember the title */
			found_games=1;
		}

		if (!strncmp(st, "PARENT", 6))
		{
			LOWER(st)

			GET_TOKEN(token, st_ptr)                /* skip name token */
			GET_TOKEN(token, st_ptr)                /* skip name token */
			HOLD_PARENT_NAME(token)                 /* Remember the parent */
			HOLD_ROMOF(token)                       /* ...and romof */
		}

		if (!strncmp(st, "DIRECTORY", 9))
		{
			LOWER(st)

			GET_TOKEN(token, st_ptr)               /* skip name token */
			GET_TOKEN(token, st_ptr)               /* skip name token */
			HOLD_GAME_NAME(token)                  /* Remember the directory */
		}

		if ((in_z80 || in_samples) && strlen(st)!=0 && st[0]!='[' && strncmp(st, "TOTAL", 5) && strncmp(st, "LOAD8", 5))
		{
			STORE_S16_ROMS(st, 0, 1)              /* Only stores if interested! */
		}

		if (!strncmp(st, "LOAD8", 5) || !strncmp(st, "LOADROM16SWAP ", 14))
		{
			LOWER(st)
			STORE_S16_ROMS(st, 1, 1)              /* Only stores if interested! */
		}

		if (!strncmp(st, "LOADROMEVENODD", 14) || !strncmp(st, "LOADROM16SWAPXOR", 16))
		{
			LOWER(st)
			STORE_S16_ROMS(st, 1, 2)              /* Only stores if interested! */
		}

		if (!strncmp(st, "[general]", 5))
	    		STORE_GAME                             /* Only stores if interested! */
	}

	if (found_games)
		STORE_GAME                             /* Only stores if interested! */

	DONE_LOAD
}

int old_load_calice_drv(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;
	int clones=(useropts->options) & GAME_AND_CLONES;
	int remove=(useropts->options) & REMOVE_CLONES;

	int verify_section=0;
	int general_passed=0;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		SKIP_WHITESPACE(st, st_ptr)                /* Some dats have indentation */

		if (st[0]=='[')
			UPPER(st)

		if (!strncmp(st, "[GENERAL]", 9))
		{
			if (general_passed)
	    			STORE_GAME    /* Only stores if interested! */
			general_passed++;
		}

		if (!strncmp(st, "[VERIFYROMS]", 12))
		{
			verify_section++;
		}
		
		if (verify_section && st[0] && st[0]!='#' && st[0]!='[')
		{
			alt_name=0; /* just to inhibit a warning */
			LOWER(st)

			if (processing_game)
			{
				/* --- Link into game's list of ROMs --- */
				if (!games[num_games].first_rom)
					games[num_games].first_rom=&roms[num_roms];
				else
					roms[num_roms-1].next_rom=&roms[num_roms];
			}

			GET_TOKEN(token, st_ptr)
			strncpy(roms[num_roms].name, token, FILE_NAME_LENGTH);
			roms[num_roms].name[FILE_NAME_LENGTH]='\0';
			
			/*GET_TOKEN(token, st_ptr)
			sprintf(roms[num_roms].size, "%ld", strtol(token+2, NULL, 16));*/

			GET_TOKEN(token, st_ptr)
			HEX_COPY(roms[num_roms].crc, token+2);
			roms[num_roms].crc[CRC_LENGTH]='\0';

			num_roms++;
		}

		if (!strncmp(st, "GAMENAME", 8))
		{
			GET_TOKEN(token, st_ptr)                /* skip name token */
			GET_TOKEN(token, st_ptr)                /* skip name token */
			HOLD_GAME_TITLE(token)                  /* Remember the title */
		}

		if (!strncmp(st, "PARENT", 6))
		{
			LOWER(st)

			GET_TOKEN(token, st_ptr)                /* skip name token */
			GET_TOKEN(token, st_ptr)                /* skip name token */
			HOLD_PARENT_NAME(token)                 /* Remember the parent */
			HOLD_ROMOF(token)                       /* ...and romof */
		}

		if (!strncmp(st, "DIRECTORY", 9))
		{
			LOWER(st)

			GET_TOKEN(token, st_ptr)               /* skip name token */
			GET_TOKEN(token, st_ptr)               /* skip name token */
			HOLD_GAME_NAME(token)                  /* Remember the directory */
		}

		if (verify_section && !strncmp(st, "[END]", 5))
		{
	    		STORE_GAME                             /* Only stores if interested! */
			general_passed=0;
		}

		if (!strncmp(st, "[", 1) && strncmp(st, "[VERIFYROMS]", 12))
			verify_section=0;
		
	}

	DONE_LOAD
}

/* --- Save --- */

int save_calice_drv(struct dat *dat_info, struct options *useropts)
{
	struct game *games=dat_info->games;
	FILE *out, *log;
	int i;
	//int j, bits;
	//long size;
	int funny_size=0, bad_size=0, bad_parent=0, warning=1;
	struct rom *curr_rom;

	/* --- Open file for writing --- */

	if (!(out=fopen(useropts->outfile, useropts->outmode)))
	{
		printf("Error opening %s\n", useropts->outfile);
		return(1);
	}

	if (!(log=fopen("datutil.log", useropts->outmode)))
	{
		fclose(out);
		printf("Error opening %s\n", "datutil.log");
		return(1);
	}

	/* --- Tell user what is going on --- */

	if (useropts->outmode[0]=='w')
		printf("Writing");
	else
		printf("Appending to");

	printf(" %s in Calice driver format...\n", useropts->outfile);

	/* --- For each game... --- */

	for (i=0; i<dat_info->num_games; i++)
	{
		/* --- Check for invalid parent --- */
		if (games[i].flags & INVALID_PARENT)
		{
			fprintf(log, "%s: specifies an invalid parent ('%s')\n", games[i].name, games[i].parent_name);
			bad_parent++;
		}

		/* --- Write its name, description, parent etc --- */

		fprintf(out, "GAMENAME    \" %s \"\n", games[i].title);
		fprintf(out, "DIRECTORY   %s\n", games[i].name);
		if (games[i].parent_name[0])
			fprintf(out, "PARENT      %s\n", games[i].parent_name);
		fprintf(out, "\n[verifyRoms]\n\n");

		/* --- Write out its ROM details --- */

		if ((curr_rom=games[i].first_rom))
		{
			while (curr_rom)
			{
				fprintf(out, "%-12s ", curr_rom->name);

				/*if (curr_rom->size[0])
				{
					size=atoi(curr_rom->size); bits=0;
					for (j=0; j < 32; j++)
					{
						if ((1<<j) & size)
							bits++;
					}
					if (bits!=1)
					{
						fprintf(log, "%s: %s has a funny size ('%s')\n", games[i].name, curr_rom->name, curr_rom->size);
						funny_size++;
					}
					fprintf(out, "0x%06lx ", atol(curr_rom->size));
				}
				else
				{
					fprintf(log, "%s: %s has no size\n", games[i].name, curr_rom->name);
					bad_size++;
				}*/

				fprintf(out, "0x%s\n", curr_rom->crc);

				curr_rom=curr_rom->next_rom;
			}
		}

		/* --- End of details for the game --- */

		fprintf(out, "\n[End]\n\n");
	}

	/* --- Close file --- */

	fclose(log);
	fclose(out);

	/* --- Display summary to user --- */

	if (bad_parent || funny_size || bad_size)
	{
		printf("\nWarnings (see datutil.log):\n\n");

		if (bad_parent)	
		{
			printf("%d) %d games have an invalid parent.\n", warning, bad_parent);
			warning++;
		}

		if (funny_size)	
		{
			printf("%d) %d ROMs have a funny size.\n", warning, funny_size);
			if (funny_size==dat_info->num_roms)
				printf("   Try changing the RomCenter DAT version to 1.92 (use a text editor)!\n");
			warning++;
		}

		if (bad_size)	
		{
			printf("%d) %d ROMs don't specify their size.\n", warning, bad_size);
			warning++;
		}
	}

	printf("\nAll done!\n");

	return(0);
}

/* --------------------------------------------------------------------------
 * ClrMamePro log
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_clrmame_log(char *lines)
{
	/* --- If the first line contains '[folder:' --- */

	if (strstr(lines, "[folder:"))
		return(1);

	return(0);
}

/* --- Load --- */

int load_clrmame_log(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;
	int clones=(useropts->options) & GAME_AND_CLONES;
	int remove=(useropts->options) & REMOVE_CLONES;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		SKIP_WHITESPACE(st, st_ptr)                /* Some dats have indentation */

		while (strchr(st, ']'))
			*strchr(st, ']')=' ';

		if (strstr(st, "[size:"))
		{
			alt_name=0; /* just to inhibit a warning */
			if (processing_game)
			{
				if (!games[num_games].first_rom)
					games[num_games].first_rom=&roms[num_roms];
				else
					roms[num_roms-1].next_rom=&roms[num_roms];
			}

			st_ptr=strstr(st, "rom:")+5;
			GET_TOKEN(token, st_ptr)
			strncpy(roms[num_roms].name, token, FILE_NAME_LENGTH);
			roms[num_roms].name[FILE_NAME_LENGTH]='\0';
			
			st_ptr=strstr(st, "CRC32:")+7;
			GET_TOKEN(token, st_ptr)
			HEX_COPY(roms[num_roms].crc, token);
			roms[num_roms].name[FILE_NAME_LENGTH]='\0';

			st_ptr=strstr(st, "size:")+6;
			GET_TOKEN(token, st_ptr)
			sprintf(roms[num_roms].size, "%s", token);
			roms[num_roms].name[FILE_NAME_LENGTH]='\0';

			num_roms++;
		}

		if (strstr(st, "parent:"))
		{
			st_ptr=strstr(st, "parent:")+8;
			GET_TOKEN(token, st_ptr)                /* skip name token */
			HOLD_PARENT_NAME(token)                 /* Remember the title */
		}

		if (strstr(st, "[folder:"))
		{
			st_ptr=strstr(st, "[folder:")+9;
			GET_TOKEN(token, st_ptr)               /* skip name token */
			HOLD_GAME_NAME(token)                  /* Remember the title */
			*(strchr(st, '[')-1)='\0';
			HOLD_GAME_TITLE(st)                     /* Remember the title */
		}

		if (!strcmp(st, ""))
	    		STORE_GAME                             /* Only stores if interested! */
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * Mimic GCS
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_mimic_gcs(char *lines)
{
	int i;

	/* --- If any of the first ten lines begin 'MACHINE :' --- */

	for (i=0; i<10; i++)
	{
		if (!strncmp(lines+BUFFER_SIZE*i, "MACHINE :", 9))
			return(1);
	}

	return(0);
}

/* --- Load --- */

int load_mimic_gcs(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;
	struct rom *rom_ptr;

	char rom_string[120];

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		RTRIM(st)

		if (!strncmp(st, "MACHINE :", 9))
		{
	    STORE_GAME                             /* Only stores if interested! */

			strtok(st,":");
			strtok(NULL,":");
			strtok(NULL,":");
			strtok(NULL,":");
			strncpy(rom_string, strtok(NULL,""), GAME_NAME_LENGTH);

			HOLD_GAME_NAME(rom_string)               /* hold the game name */
		}

		if (!strncmp(st, "ROMBANK :", 9) ||
			!strncmp(st, "PAL-ROM :", 9) ||
			!strncmp(st, "PAL-GLU :", 9) ||
			!strncmp(st, "PAL-SLU :", 9) ||
			!strncmp(st, "GFX-ROM :", 9))
		{
			strcpy(rom_string, strrchr(st, ':')+1);
			if (strcmp(rom_string, "NONE") && strcmp(rom_string, "ARAM"))
				STORE_ROM_UNIQUE(rom_string)                  /* Only stores if interested! */
		}

		if (!strncmp(st, "MEMMAP  :", 9))
		{
			if (!strncmp(strrchr(st, ':')-3, "ROM", 3))
			{
				strcpy(rom_string, strrchr(st, ':')+1);
				STORE_ROM_UNIQUE(rom_string)                  /* Only stores if interested! */
			}
		}
	}

	STORE_GAME                                 /* Only stores if interested! */

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * RomCenter Dat - Load and Save
 *
 * The loader doesn't use macros like everything else as RomCenter dats are so
 * different in nature to the others it has it's own algorithm!
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_romcenter_dat(char *lines)
{
	int i, found=0;

	/* --- If any of first ten lines begin with '[CREDITS]' --- */

	for (i=0; i<20; i++)
		if (!strncmp(lines+i*BUFFER_SIZE, "[CREDITS]", 9))
			found++;

	if (found)
		for (i=0; i<20; i++)
			if (!strncmp(lines+i*BUFFER_SIZE, "version=1", 9))
				return(1);
		
	return(0);
}

/* --- Load --- */

int load_romcenter_dat(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;
	int clones=(useropts->options) & GAME_AND_CLONES;
	int remove=(useropts->options) & REMOVE_CLONES;

	struct rom *next_rom;

	char parent_name[GAME_NAME_LENGTH+1];
	char game_name[GAME_NAME_LENGTH+1];
	char game_title[TITLE_LENGTH+1];
	char rom_name[FILE_NAME_LENGTH+1];
	long flags;
	char merge_name[FILE_NAME_LENGTH+1];
	char rom_size[SIZE_LENGTH+1];
	char rom_crc[CRC_LENGTH+1];

	char *ptr;

	int curr_game=0;
	int found_game;
	int emulator_header=0;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		if (st[0]=='[')
			emulator_header=0;

		if (!strcmp(st, "[CREDITS]"))
			emulator_header=1;

		if (!strcmp(st, "[EMULATOR]"))
			emulator_header=2;

		if (emulator_header==1 && !strncmp(st, "Author", 6) && !(useropts->rc_author))
		{
			if (strchr(st, '='))
			{
				strcpy(author, strchr(st, '=')+1);
				useropts->rc_author=author;
			}
		}

		if (emulator_header==1 && !strncmp(st, "Version", 7) && !(useropts->rc_version))
		{
			if (strchr(st, '='))
			{
				strcpy(version, strchr(st, '=')+1);
				useropts->rc_version=version;
			}
		}

		if (emulator_header==1 && !strncmp(st, "Comment", 7) && !(useropts->rc_comment))
		{
			if (strchr(st, '='))
			{
				strcpy(comment, strchr(st, '=')+1);
				useropts->rc_comment=comment;
			}
		}

		if (emulator_header==2 && !strncmp(st, "refname", 7) && !(useropts->rc_refname))
		{
			if (strchr(st, '='))
			{
				strcpy(refname, strchr(st, '=')+1);
				useropts->rc_refname=refname;
			}
		}

		if (emulator_header==2 && !strncmp(st, "version", 7) && !(useropts->rc_fullname))
		{
			if (strchr(st, '='))
			{
				strcpy(fullname, strchr(st, '=')+1);
				useropts->rc_fullname=fullname;
			}
		}

		/* --- We're only interested in ROM lines --- */

		if (st[0]=='')
		{
			/* --- Extract parent name and game name --- */

			strncpy(parent_name, strtok(st,""), GAME_NAME_LENGTH);
			parent_name[GAME_NAME_LENGTH]='\0';
			strtok(NULL,"");
			strncpy(game_name, strtok(NULL,""), GAME_NAME_LENGTH);
			game_name[GAME_NAME_LENGTH]='\0';

			/* --- Are we interested in this line? --- */
			/* --- Dependant on -g and -c options --- */

			if (((!game)||
					(game && !strcmp(game, game_name)) ||
					(game && clones && !strcmp(game, parent_name))) &&
					((!remove) || (remove && !strcmp(game_name, parent_name))))
			{
				/* --- Hard coded logic to say neogeo is a resource --- */

				if (!strcmp(game_name, "neogeo") || !strcmp(game_name, "playch10") || !strcmp(game_name, "cvs") || !strcmp(game_name, "decocass"))
					flags=RESOURCE;
				else
					flags=0;

				/* --- Extract other details about game and ROM --- */

				strncpy(game_title, strtok(NULL,""), TITLE_LENGTH);
				game_title[TITLE_LENGTH]='\0';
				strncpy(rom_name, strtok(NULL,""), FILE_NAME_LENGTH);
				rom_name[FILE_NAME_LENGTH]='\0';
				strcpy(rom_crc, strtok(NULL,""));
				if (strcmp(strtok(NULL,""), "N"))
					flags|=NEOGEO_GAME;

				/* --- Best we can do for merge name --- */

				if (!strcmp(strtok(NULL,""), "N"))
					sprintf(merge_name, rom_name);
				else
					merge_name[0]='\0';

				/* --- Size may or may not be there --- */

				if ((ptr=strtok(NULL,"")))
					strcpy(rom_size, ptr);
				else
					rom_size[0]='\0';

				/* --- Store the ROM details --- */

				if (strcmp(rom_crc, "        "))
				{
					strcpy(roms[num_roms].name, rom_name);
					strcpy(roms[num_roms].merge, merge_name);
					strcpy(roms[num_roms].size, rom_size);
					strcpy(roms[num_roms].crc, rom_crc);
				}

				/* --- See if the game details are already stored --- */

				found_game=0;
				for (curr_game=num_games-1; !found_game && curr_game>=0; curr_game--)
				{
					if (!strcmp(game_name, games[curr_game].name))
						found_game=1;
				}

				/* --- If not... --- */

				if (!found_game)
				{
					/* --- Store game details --- */

					strcpy(games[num_games].name, game_name);
					strcpy(games[num_games].title, game_title);
					if (strcmp(game_name, parent_name))
					{
						strcpy(games[num_games].parent_name, parent_name);
						num_clones++;
					}

					/* --- If game has parent, assume 'romof' is the same --- */

					if (strcmp(parent_name, game_name))
						strcpy(games[num_games].romof, parent_name);
					else
						games[num_games].romof[0]='\0';

					/* --- Store the flags --- */

					games[num_games].flags=flags;

					/* --- If game is for Neo-Geo and romof not specified... --- */
					if ((games[num_games].flags & NEOGEO_GAME) &&
							strcmp(games[num_games].name, "neogeo") &&
							!games[num_games].romof[0])
						sprintf(games[num_games].romof, "neogeo");

					/* --- Let the game know where the first ROM is stored --- */

					if (strcmp(rom_crc, "        "))
						games[num_games].first_rom=&roms[num_roms];

					/* --- Increment game and resource counters --- */

					num_games++;

					if (flags & RESOURCE)
						num_resources++;
				}
				else
				{
					/* --- We found the game but have gone one past it --- */

					curr_game++;

					/* --- If it doesn't have any ROMs (actually impossible) --- */

					if (strcmp(rom_crc, "        "))
					{
						if (!games[curr_game].first_rom)
							games[curr_game].first_rom=&roms[num_roms];
						else
						{
							/* --- Walk through list until last ROM reached --- */

							next_rom=games[curr_game].first_rom;

							while (next_rom->next_rom)
								next_rom=next_rom->next_rom;

							/* --- Tell it where the last one is --- */
	
							next_rom->next_rom=&roms[num_roms];
						}
					}
				}

				/* --- One more ROM (and it may have been for a resource) --- */

				if (strcmp(rom_crc, "        "))
				{
					num_roms++;

					if (flags & RESOURCE)
						num_resource_roms++;
				}
			}
		}
	}

	DONE_LOAD
}

/* --- Save --- */

int save_romcenter_dat(struct dat *dat_info, struct options *useropts)
{
	struct game *games=dat_info->games;
	FILE *out, *log;
	int i, j, bits;
	long size;
	int bad_parent=0, funny_size=0, bad_size=0, bad_merge=0, bad_romof=0, warning=1;
	struct rom *curr_rom;

	/* --- Open file for writing --- */

	if (!(out=fopen(useropts->outfile, useropts->outmode)))
	{
		printf("Error opening %s\n", useropts->outfile);
		return(1);
	}

	if (!(log=fopen("datutil.log", useropts->outmode)))
	{
		fclose(out);
		printf("Error opening %s\n", "datutil.log");
		return(1);
	}

	/* --- Tell user what is going on and write header if required --- */

	if (useropts->outmode[0]=='w')
	{
		printf("Writing");
		fprintf(out, "[CREDITS]\n");

		if (!(useropts->rc_author))
			fprintf(out, "Author=\n");
		else
			fprintf(out, "Author=%s\n", useropts->rc_author);

		if (!(useropts->rc_version))
			fprintf(out, "Version=\n");
		else
			fprintf(out, "Version=%s\n", useropts->rc_version);

		if (!(useropts->rc_comment))
			fprintf(out, "Comment=Generated by DatUtil %s\n", DATUTIL_VERSION);
		else
			fprintf(out, "Comment=%s\n", useropts->rc_comment);

		fprintf(out, "[DAT]\n");
		fprintf(out, "version=1.90\n");
		fprintf(out, "[EMULATOR]\n");

		if (!(useropts->rc_refname))
			fprintf(out, "refname=\n");
		else
			fprintf(out, "refname=%s\n", useropts->rc_refname);

		if (!(useropts->rc_fullname))
			fprintf(out, "version=\n");
		else
			fprintf(out, "version=%s\n", useropts->rc_fullname);

		fprintf(out, "exe=\n");
		fprintf(out, "runcmd=\n");
		fprintf(out, "romspaths=\n");

		fprintf(out, "[GAMES]\n");
	}
	else
		printf("Appending to");

	printf(" %s in RomCenter format...\n", useropts->outfile);

	/* --- For each game --- */

	for (i=0; i<dat_info->num_games; i++)
	{
		/* --- Check for invalid parent --- */
		if (games[i].flags & INVALID_PARENT)
		{
			fprintf(log, "%s: specifies an invalid parent ('%s')\n", games[i].name, games[i].parent_name);
			bad_parent++;
		}

		/* --- If 'romof' differs from parent, we can't represent this --- */

		if (strcmp(games[i].parent_name, games[i].romof) &&
				strcmp(games[i].romof, "neogeo") &&
				strcmp(games[i].romof, "playch10") &&
				strcmp(games[i].romof, "cvs") &&
				strcmp(games[i].romof, "decocass"))
		{
			fprintf(log, "%s: specifies different cloneof and romof ('%s' and '%s')\n",
				games[i].name, games[i].parent_name, games[i].romof);
			bad_romof++;
		}

		/* --- If the game doesn't have a parent, then use the game itself --- */

		if (!games[i].parent_name[0])
		{
			strcpy(games[i].parent_name, games[i].name);
			strcpy(games[i].parent_title, games[i].title);
		}

		/* --- Write out the ROM details --- */

		if ((curr_rom=games[i].first_rom))
		{
			while (curr_rom)
			{
				/* --- Don't keep Neo-Geo ROMs (only in the neogeo resource) --- */

				if ((games[i].flags & RESOURCE) || !is_resource_rom(curr_rom))
				{
					/* --- General details --- */

					fprintf(out, "%s%s%s%s%s%s", games[i].parent_name, games[i].parent_title, games[i].name, games[i].title, curr_rom->name, curr_rom->crc);

					/* --- Neo-Geo flag and merge flag --- */

					if (games[i].flags & NEOGEO_GAME)
						fprintf(out, "O");
					else
						fprintf(out, "N");

					if (!curr_rom->merge[0])
						fprintf(out, "O");
					else
					{	
						/* --- If merge name differs, we can't represent this --- */

						if (strcmp(curr_rom->merge, curr_rom->name))
						{
							fprintf(log, "%s: %s has a different merge name ('%s')\n", games[i].name, curr_rom->name, curr_rom->merge);
							bad_merge++;
						}

						fprintf(out, "N");
					}

					/* --- The ROM size --- */

					if (curr_rom->size[0])
					{
						size=atoi(curr_rom->size); bits=0;
						for (j=0; j < 32; j++)
						{
							if ((1<<j) & size)
								bits++;
						}
						if (bits!=1)
						{
							fprintf(log, "%s: %s has a funny size ('%s')\n", games[i].name, curr_rom->name, curr_rom->size);
							funny_size++;
						}
						fprintf(out, "%s", curr_rom->size);
					}
					else
					{
						fprintf(log, "%s: %s has no size\n", games[i].name, curr_rom->name);
						bad_size++;
					}

					fprintf(out, "\n");
				}

				curr_rom=curr_rom->next_rom;
			}
		}
		else
		{
			fprintf(out, "%s%s%s%sNO_ROMS        NO\n", games[i].parent_name, games[i].parent_title, games[i].name, games[i].title);
		}
	}

	/* --- Close output files --- */

	fclose(log);
	fclose(out);

	/* --- Display summary to user --- */

	if (bad_parent || funny_size || bad_size || bad_merge || bad_romof || dat_info->num_samples)
	{
		printf("\nWarnings (see datutil.log):\n\n");

		if (bad_parent)	
		{
			printf("%d) %d games have an invalid parent.\n", warning, bad_parent);
			warning++;
		}

		if (funny_size)	
		{
			printf("%d) %d ROMs have a funny size.\n", warning, funny_size);
			if (funny_size==dat_info->num_roms)
				printf("   Try changing the RomCenter DAT version to 1.92 (use a text editor)!\n");
			warning++;
		}

		if (bad_size)	
		{
			printf("%d) %d ROMs don't specify their size.\n", warning, bad_size);
			warning++;
		}

		if (bad_merge)	
		{
			printf("%d) %d ROMs specify merge name different to their own.\n   This information cannot be represented in RomCenter datafiles.\n", warning, bad_merge);
			warning++;
		}

		if (bad_romof)	
		{
			printf("%d) %d games specify romof different to their parent.\n   This information cannot be represented in RomCenter datafiles.\n", warning, bad_romof);
			warning++;
		}

		if (dat_info->num_samples)	
		{
			printf("%d) %d samples were listed in the original file.\n   This information cannot be represented in RomCenter datafiles.\n", warning, dat_info->num_samples);
			warning++;
		}
	}

	printf("\nAll done!\n");

	return(0);
}


/* --------------------------------------------------------------------------
 * RomCenter 2 Dat - Load and Save
 *
 * The loader doesn't use macros like everything else as RomCenter dats are so
 * different in nature to the others it has it's own algorithm!
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_romcenter2_dat(char *lines)
{
	int i, found=0;

	/* --- If any of first ten lines begin with '[CREDITS]' --- */

	for (i=0; i<20; i++)
		if (!strncmp(lines+i*BUFFER_SIZE, "[CREDITS]", 9))
			found++;

	if (found)
		for (i=0; i<20; i++)
			if (!strncmp(lines+i*BUFFER_SIZE, "version=2", 9))
				return(1);
		
	return(0);
}

/* --- Load --- */

int load_romcenter2_dat(struct dat *dat_info, struct options *useropts)
{
	char *game=useropts->game;
	int clones=(useropts->options) & GAME_AND_CLONES;
	int remove=(useropts->options) & REMOVE_CLONES;

	struct rom *next_rom;

	char parent_name[GAME_NAME_LENGTH+1];
	char game_name[GAME_NAME_LENGTH+1];
	char game_title[TITLE_LENGTH+1];
	char romof_name[GAME_NAME_LENGTH+1];
	char rom_name[FILE_NAME_LENGTH+1];
	long flags;
	char merge_name[FILE_NAME_LENGTH+1];
	char rom_size[SIZE_LENGTH+1];
	char rom_crc[CRC_LENGTH+1];

	char *ptr;

	int curr_game=0;
	int found_game;
	int emulator_header=0;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		if (st[0]=='[')
			emulator_header=0;

		if (!strcmp(st, "[CREDITS]"))
			emulator_header=1;

		if (!strcmp(st, "[EMULATOR]"))
			emulator_header=2;

		if (emulator_header==1 && !strncmp(st, "Author", 6) && !(useropts->rc_author))
		{
			if (strchr(st, '='))
			{
				strcpy(author, strchr(st, '=')+1);
				useropts->rc_author=author;
			}
		}

		if (emulator_header==1 && !strncmp(st, "Version", 7) && !(useropts->rc_version))
		{
			if (strchr(st, '='))
			{
				strcpy(version, strchr(st, '=')+1);
				useropts->rc_version=version;
			}
		}

		if (emulator_header==1 && !strncmp(st, "Comment", 7) && !(useropts->rc_comment))
		{
			if (strchr(st, '='))
			{
				strcpy(comment, strchr(st, '=')+1);
				useropts->rc_comment=comment;
			}
		}

		if (emulator_header==2 && !strncmp(st, "refname", 7) && !(useropts->rc_refname))
		{
			if (strchr(st, '='))
			{
				strcpy(refname, strchr(st, '=')+1);
				useropts->rc_refname=refname;
			}
		}

		if (emulator_header==2 && !strncmp(st, "version", 7) && !(useropts->rc_fullname))
		{
			if (strchr(st, '='))
			{
				strcpy(fullname, strchr(st, '=')+1);
				useropts->rc_fullname=fullname;
			}
		}

		/* --- We're only interested in ROM lines --- */

		if (st[0]=='')
		{
			/* --- Extract parent name and game name --- */

			strncpy(parent_name, strtok(st,""), GAME_NAME_LENGTH);
			parent_name[GAME_NAME_LENGTH]='\0';
			strtok(NULL,"");
			strncpy(game_name, strtok(NULL,""), GAME_NAME_LENGTH);
			game_name[GAME_NAME_LENGTH]='\0';

			/* --- Are we interested in this line? --- */
			/* --- Dependant on -g and -c options --- */

			if (((!game)||
					(game && !strcmp(game, game_name)) ||
					(game && clones && !strcmp(game, parent_name))) &&
					((!remove) || (remove && !strcmp(game_name, parent_name))))
			{
				/* --- Hard coded logic to say neogeo is a resource --- */

				if (!strcmp(game_name, "neogeo") || !strcmp(game_name, "playch10") || !strcmp(game_name, "cvs") || !strcmp(game_name, "decocass"))
					flags=RESOURCE;
				else
					flags=0;

				/* --- Extract other details about game and ROM --- */

				strncpy(game_title, strtok(NULL,""), TITLE_LENGTH);
				game_title[TITLE_LENGTH]='\0';
				strncpy(rom_name, strtok(NULL,""), FILE_NAME_LENGTH);
				rom_name[FILE_NAME_LENGTH]='\0';
				strcpy(rom_crc, strtok(NULL,""));
				ptr=strtok(NULL,"");
				if (ptr)
				{
					strncpy(rom_size, ptr, SIZE_LENGTH);
					rom_size[SIZE_LENGTH]='\0';
				}
				else
				{
					strcpy(rom_size, "0");
				}
				ptr=strtok(NULL,"");
				if (ptr)
				{
					strncpy(romof_name, ptr, GAME_NAME_LENGTH);
					romof_name[GAME_NAME_LENGTH]='\0';
				}
				else
				{
					if (strcmp(parent_name, game_name))
						strcpy(romof_name, parent_name);
					else
						*romof_name='\0';
				}
				ptr=strtok(NULL,"");
				if (ptr)
				{
					strncpy(merge_name, ptr, FILE_NAME_LENGTH);
					merge_name[FILE_NAME_LENGTH]='\0';
				}
				else
					*merge_name='\0';

				/* --- Store the ROM details --- */

				if (strcmp(rom_crc, "        "))
				{
					strcpy(roms[num_roms].name, rom_name);
					strcpy(roms[num_roms].merge, merge_name);
					strcpy(roms[num_roms].size, rom_size);
					strcpy(roms[num_roms].crc, rom_crc);
				}

				/* --- See if the game details are already stored --- */

				found_game=0;
				for (curr_game=num_games-1; !found_game && curr_game>=0; curr_game--)
				{
					if (!strcmp(game_name, games[curr_game].name))
						found_game=1;
				}

				/* --- If not... --- */

				if (!found_game)
				{
					/* --- Store game details --- */

					strcpy(games[num_games].name, game_name);
					strcpy(games[num_games].title, game_title);
					if (strcmp(game_name, parent_name))
					{
						strcpy(games[num_games].parent_name, parent_name);
						num_clones++;
					}

					/* --- If game has parent, assume 'romof' is the same --- */

					strcpy(games[num_games].romof, romof_name);

					/* --- Store the flags --- */

					games[num_games].flags=flags;

					/* --- Let the game know where the first ROM is stored --- */

					if (strcmp(rom_crc, "        "))
						games[num_games].first_rom=&roms[num_roms];

					/* --- Increment game and resource counters --- */

					num_games++;

					if (flags & RESOURCE)
						num_resources++;
				}
				else
				{
					/* --- We found the game but have gone one past it --- */

					curr_game++;

					/* --- If it doesn't have any ROMs (actually impossible) --- */

					if (strcmp(rom_crc, "        "))
					{
						if (!games[curr_game].first_rom)
							games[curr_game].first_rom=&roms[num_roms];
						else
						{
							/* --- Walk through list until last ROM reached --- */

							next_rom=games[curr_game].first_rom;

							while (next_rom->next_rom)
								next_rom=next_rom->next_rom;

							/* --- Tell it where the last one is --- */
	
							next_rom->next_rom=&roms[num_roms];
						}
					}
				}

				/* --- One more ROM (and it may have been for a resource) --- */

				if (strcmp(rom_crc, "        "))
				{
					num_roms++;

					if (flags & RESOURCE)
						num_resource_roms++;
				}
			}
		}
	}

	DONE_LOAD
}

/* --- Save --- */

int save_romcenter2_dat(struct dat *dat_info, struct options *useropts)
{
	struct game *games=dat_info->games;
	FILE *out, *log;
	int i, j, bits;
	long size;
	int bad_parent=0, funny_size=0, bad_size=0, warning=1;
	struct rom *curr_rom;

	/* --- Open file for writing --- */

	if (!(out=fopen(useropts->outfile, useropts->outmode)))
	{
		printf("Error opening %s\n", useropts->outfile);
		return(1);
	}

	if (!(log=fopen("datutil.log", useropts->outmode)))
	{
		fclose(out);
		printf("Error opening %s\n", "datutil.log");
		return(1);
	}

	/* --- Tell user what is going on and write header if required --- */

	if (useropts->outmode[0]=='w')
	{
		printf("Writing");
		fprintf(out, "[CREDITS]\n");
		if (!(useropts->rc_author))
			fprintf(out, "Author=\n");
		else
			fprintf(out, "Author=%s\n", useropts->rc_author);

		if (!(useropts->rc_version))
			fprintf(out, "Version=\n");
		else
			fprintf(out, "Version=%s\n", useropts->rc_version);

		if (!(useropts->rc_comment))
			fprintf(out, "Comment=Generated by DatUtil %s\n", DATUTIL_VERSION);
		else
			fprintf(out, "Comment=%s\n", useropts->rc_comment);

		fprintf(out, "[DAT]\n");
		fprintf(out, "version=2.00\n");
		fprintf(out, "[EMULATOR]\n");

		if (!(useropts->rc_refname))
			fprintf(out, "refname=\n");
		else
			fprintf(out, "refname=%s\n", useropts->rc_refname);

		if (!(useropts->rc_fullname))
			fprintf(out, "version=\n");
		else
			fprintf(out, "version=%s\n", useropts->rc_fullname);

		fprintf(out, "[GAMES]\n");
	}
	else
		printf("Appending to");

	printf(" %s in RomCenter 2 format...\n", useropts->outfile);

	/* --- For each game --- */

	for (i=0; i<dat_info->num_games; i++)
	{
		/* --- Check for invalid parent --- */
		if (games[i].flags & INVALID_PARENT)
		{
			fprintf(log, "%s: specifies an invalid parent ('%s')\n", games[i].name, games[i].parent_name);
			bad_parent++;
		}

		/* --- If the game doesn't have a parent, then use the game itself --- */

		if (!games[i].parent_name[0])
		{
			strcpy(games[i].parent_name, games[i].name);
			strcpy(games[i].parent_title, games[i].title);
		}

		/* --- Write out the ROM details --- */

		if ((curr_rom=games[i].first_rom))
		{
			while (curr_rom)
			{
				/* --- General details --- */

				if (strcmp(games[i].name, games[i].parent_name) || (games[i].romof[0] && strcmp(games[i].name, games[i].romof)))
					fprintf(out, "%s%s%s%s%s%s%s%s%s\n", games[i].parent_name, games[i].parent_title, games[i].name, games[i].title, curr_rom->name, curr_rom->crc, curr_rom->size, games[i].romof, curr_rom->merge);
				else
					fprintf(out, "%s%s%s%s%s%s%s%s\n", games[i].parent_name, games[i].parent_title, games[i].name, games[i].title, curr_rom->name, curr_rom->crc, curr_rom->size, games[i].romof);

				/* --- The ROM size --- */

				if (curr_rom->size[0])
				{
					size=atoi(curr_rom->size); bits=0;
					for (j=0; j < 32; j++)
					{
						if ((1<<j) & size)
							bits++;
					}
					if (bits!=1)
					{
						fprintf(log, "%s: %s has a funny size ('%s')\n", games[i].name, curr_rom->name, curr_rom->size);
						funny_size++;
					}
				}
				else
				{
					fprintf(log, "%s: %s has no size\n", games[i].name, curr_rom->name);
					bad_size++;
				}

				curr_rom=curr_rom->next_rom;
			}
		}
		else
		{
			fprintf(out, "%s%s%s%sNO_ROMS        0\n", games[i].parent_name, games[i].parent_title, games[i].name, games[i].title);
		}
	}

	/* --- Close output files --- */

	fclose(log);
	fclose(out);

	/* --- Display summary to user --- */

	if (bad_parent || funny_size || bad_size || dat_info->num_samples)
	{
		printf("\nWarnings (see datutil.log):\n\n");

		if (bad_parent)	
		{
			printf("%d) %d games have an invalid parent.\n", warning, bad_parent);
			warning++;
		}

		if (funny_size)	
		{
			printf("%d) %d ROMs have a funny size.\n", warning, funny_size);
			if (funny_size==dat_info->num_roms)
				printf("   Try changing the RomCenter DAT version to 1.92 (use a text editor)!\n");
			warning++;
		}

		if (bad_size)	
		{
			printf("%d) %d ROMs don't specify their size.\n", warning, bad_size);
			warning++;
		}

		if (dat_info->num_samples)	
		{
			printf("%d) %d samples were listed in the original file.\n   This information cannot be represented in RomCenter 2 datafiles.\n", warning, dat_info->num_samples);
			warning++;
		}
	}

	printf("\nAll done!\n");

	return(0);
}

/* --------------------------------------------------------------------------
 * Directory scan - Load only
 *
 * Scans a directory which contains either ZIP files or directories
 * -------------------------------------------------------------------------- */

int dir_scan(struct dat *dat_info, struct options *useropts)
{
	DIR *dirp, *sdirp;                                    
	struct dirent *direntp, *sdirentp;                       
	struct stat buf;

	ZIP *zip;
	struct zipent *zipent;

	char fn[BUFFER_SIZE+1];
	char *game=useropts->game;
	unsigned long crc;

	INIT_LOAD

	dirp = opendir(dat_info->name);                          

	while (dirp && ((direntp = readdir(dirp)) != NULL))
	{
		sprintf(fn, "%s/%s", dat_info->name, direntp->d_name);
		if (stat(fn, &buf) == 0)
		{
			if (!(buf.st_mode & S_IFDIR))
			{
				strcpy(st, fn);
				UPPER(st)
				if (strrchr(st, '.') && !strcmp(strrchr(st, '.'), ".ZIP"))
				{
					strcpy(st, direntp->d_name);
					if (strstr(st, ".zip"))
						*strstr(st, ".zip")='\0';
					HOLD_GAME_NAME(st)  /* Remember the game name */
					HOLD_GAME_TITLE(st)  /* Remember the game name */

					if ((zip=openzip(fn)))
					{
						while ((zipent = readzip(zip)) != 0)
						{
							if (zipent->uncompressed_size)
							{
							sprintf(st, "rom ( name %s size %d crc %08lx )",
								zipent->name, zipent->uncompressed_size, (unsigned long)zipent->crc32);
							STORE_ROM(st)  /* Only stores if interested! */
							}
						}
						closezip(zip);
					}

					STORE_GAME
				}
			}
			else
			{
				if (fn[strlen(fn)-1]!='.')	/* Don't try . or .. entries */
				{
					HOLD_GAME_NAME(direntp->d_name)  /* Remember the game name */
					HOLD_GAME_TITLE(direntp->d_name)  /* Remember the game name */

					sdirp = opendir(fn);                          

					while (sdirp && ((sdirentp = readdir(sdirp)) != NULL))
					{
						sprintf(st, "%s/%s", fn, sdirentp->d_name);
						if (stat(st, &buf) == 0)
						{
							if (!(buf.st_mode & S_IFDIR) && buf.st_size)
							{
									crc=CalcFileCRC32(st, buf.st_size);
				
									sprintf(st, "rom ( name %s size %d crc %08lx )",
										sdirentp->d_name, (int)buf.st_size, crc);
									STORE_ROM(st)  /* Only stores if interested! */
							}
						}
						else
						{
							printf("Error getting attributes of %s\n", sdirentp->d_name);
						}
					}

					if (sdirp)
						closedir(sdirp);                               

					STORE_GAME
				}
			}
		}
		else
		{
			printf("Error getting attributes of %s\n", direntp->d_name);
		}
	}

	if (dirp)
		closedir(dirp);                               

	DONE_LOAD
}                                                  

/* --------------------------------------------------------------------------
 * CRC calculation uses zlib function
 * -------------------------------------------------------------------------- */

unsigned long CalcFileCRC32(char *fn, unsigned long reclen)
{
	FILE *in;
	unsigned char *mem;
	unsigned long crc=0;

	if (!(in=fopen(fn, "rb")))
	{
		printf("Error opening %s\n", fn);
		return(0);
	}

	if ((mem=malloc(reclen)))
	{
		fread(mem, 1, reclen, in);
		crc = crc32(0, NULL, 0);
		crc = crc32(crc, mem, reclen);
		free(mem);
	}

	fclose(in);

	return(crc);
}
